import ase.io
import numpy as np
import pytest
from ase.build import molecule
from ase.calculators.vdwcorrection import vdWTkatchenko09prl
from ase.parallel import barrier, parprint

from gpaw import GPAW
from gpaw.analyse.hirshfeld import HirshfeldPartitioning
from gpaw.analyse.vdwradii import vdWradii
from gpaw.cluster import Cluster
from gpaw.test import equal


@pytest.mark.later
def test_vdw_ts09(in_tmp_dir):
    h = 0.4
    s = Cluster(molecule('LiH'))
    s.minimal_box(3., h=h)

    def print_charge_and_check(hp, q=0, label='unpolarized'):
        q_a = np.array(hp.get_charges())
        parprint('Charges ({0})='.format(label), q_a, ', sum=', q_a.sum())
        equal(q_a.sum(), q, 0.03)
        return q_a

    # spin unpolarized

    if 1:
        out_traj = 'LiH.traj'
        out_txt = 'LiH.txt'

        cc = GPAW(mode='fd', h=h, xc='PBE', txt=out_txt)

        # this is needed to initialize txt output
        cc.initialize(s)

        hp = HirshfeldPartitioning(cc)
        c = vdWTkatchenko09prl(hp,
                               vdWradii(s.get_chemical_symbols(), 'PBE'))
        s.calc = c
        E = s.get_potential_energy()
        F_ac = s.get_forces()
        s.write(out_traj)
        q_a = print_charge_and_check(hp)

        barrier()

        # test I/O, accuracy due to text output
        accuracy = 1.e-5
        for fname in [out_traj, out_txt]:
            s_out = ase.io.read(fname)
            equal(s_out.get_potential_energy(), E, accuracy)
            for fi, fo in zip(F_ac, s_out.get_forces()):
                equal(fi, fo, accuracy)

    # spin polarized

    if 0:
        ccs = GPAW(mode='fd', h=h, xc='PBE', spinpol=True,
                   txt=None)
        hps = HirshfeldPartitioning(ccs)
        cs = vdWTkatchenko09prl(hps, vdWradii(s.get_chemical_symbols(), 'PBE'))
        s.calc = cs
        Es = s.get_potential_energy()
        Fs_ac = s.get_forces()

        qs_a = print_charge_and_check(hps, label='spin')

        equal(q_a, qs_a, 1.e-6)
        equal(E, Es, 1.e-4)
        equal(F_ac, Fs_ac, 1.e-4)

    # charged

    if 0:
        cc = cc.new(charge=1)
        hpp = HirshfeldPartitioning(cc)
        cp = vdWTkatchenko09prl(hpp,
                                vdWradii(s.get_chemical_symbols(), 'PBE'))
        s.calc = cp
        E = s.get_potential_energy()

        print_charge_and_check(hpp, 1, label='+1')



import pytest
from gpaw.mpi import world
from ase import Atom, Atoms
from gpaw import GPAW
from gpaw.test import equal

pytestmark = pytest.mark.skipif(world.size > 1,
                                reason='world.size > 1')


def test_pw_davidson_pw():
    a = 4.05
    d = a / 2**0.5
    bulk = Atoms([Atom('Al', (0, 0, 0)),
                  Atom('Al', (0.5, 0.5, 0.5))], pbc=True)
    bulk.set_cell((d, d, a), scale_atoms=True)
    calc = GPAW(mode='pw',
                nbands=2 * 8,
                kpts=(2, 2, 2),
                convergence={'eigenstates': 7.2e-9, 'energy': 1e-5})
    bulk.calc = calc
    e0 = bulk.get_potential_energy()
    niter0 = calc.get_number_of_iterations()
    calc = GPAW(mode='pw',
                nbands=2 * 8,
                kpts=(2, 2, 2),
                convergence={'eigenstates': 7.2e-9,
                             'energy': 1e-5,
                             'bands': 5},
                eigensolver='dav')
    bulk.calc = calc
    e1 = bulk.get_potential_energy()
    niter1 = calc.get_number_of_iterations()
    equal(e0, e1, 5.0e-6)

    energy_tolerance = 0.0004
    equal(e0, -6.97798, energy_tolerance)
    assert 8 <= niter0 <= 12, niter0
    equal(e1, -6.97798, energy_tolerance)
    assert 8 <= niter1 <= 22, niter1



from ase import Atoms
from gpaw import GPAW
from gpaw.mpi import rank, size


def test_parallel_compare():
    a = 3.0
    H = Atoms('H',
              cell=(a, a, a),
              pbc=True,
              calculator=GPAW(mode='fd'))
    if size > 1:
        H.positions[0, 0] += 0.01 * rank
        try:
            H.get_potential_energy()
        except ValueError as e:
            err_ranks = e.args[1]
            assert (err_ranks == range(1, size)).all()
        else:
            assert False



from ase import Atom, Atoms
from gpaw import GPAW
from gpaw.test import equal


def test_generic_2Al():
    a = 4.05
    d = a / 2**0.5
    bulk = Atoms([Atom('Al', (0, 0, 0)),
                  Atom('Al', (0, 0, d))],
                 cell=(4 * d, 4 * d, 2 * d),
                 pbc=1)
    n = 16
    calc = GPAW(mode='fd',
                gpts=(2 * n, 2 * n, 1 * n),
                nbands=1 * 8,
                kpts=(1, 1, 4),
                convergence={'eigenstates': 2.3e-9})
    bulk.calc = calc
    e2 = bulk.get_potential_energy()

    bulk = bulk.repeat((1, 1, 2))
    calc = GPAW(mode='fd',
                gpts=(2 * n, 2 * n, 2 * n),
                nbands=16,
                kpts=(1, 1, 2),
                convergence={'eigenstates': 2.3e-9})
    bulk.calc = calc
    e4 = bulk.get_potential_energy()

    # checks
    energy_tolerance = 0.002

    print(e2, e4)
    equal(e4 / 2, e2, 48e-6)
    equal(e2, -3.41595, energy_tolerance)
    equal(e4, -6.83191, energy_tolerance)



import pytest
import numpy as np

from ase.io.ulm import ulmopen
from gpaw.test.conftest import response_band_cutoff


@pytest.mark.response
@pytest.mark.later  # new code doesn't understand interpolation keyword
@pytest.mark.parametrize('gs', list(response_band_cutoff.keys()))
def test_response_band_cutoff(in_tmp_dir, gpw_files, gs):

    nbands = response_band_cutoff[gs]
    with ulmopen(gpw_files[gs]) as reader:
        eps_skn = reader.wave_functions.eigenvalues
        nconv = reader.parameters.convergence['bands']
    print(gs)
    assert nbands < nconv
    possible_cutoffs = get_nbands_cutoff_list(eps_skn, nconv)
    print(possible_cutoffs)
    assert nbands in possible_cutoffs


def get_nbands_cutoff_list(eps_skn, nconv, atol=1e-3):
    """ Possible cutoffs for response calc
    Returns the set all allowed band cutoffs in a response calculation.
    Assures that there are no  degeneracies at the edge of the cutoff

    >>> eps = np.array([[[0, 1, 2, 2, 3, 4, 5, 5]]])
    >>> get_nbands_cutoff_list(eps, 7)
    {1, 2, 4, 5, 6}
    """
    allset = set(range(nconv + 1))
    # Loop over spins and k-points
    for eps_kn in eps_skn:
        for eps_n in eps_kn:
            # check degenerate eigenvalues
            cutlist = np.isclose(eps_n[:nconv - 1], eps_n[1:nconv],
                                 atol=atol)
            cutlist = np.argwhere(~cutlist)
            # cutoff is allowed index + 1
            cutlist += 1
            thisset = set(cutlist.flatten())
            # find minimum cutoff that works for all k
            allset = thisset & allset

    return allset



from __future__ import annotations

from typing import Generator

import numpy as np
from ase.dft.bandgap import bandgap
from ase.io.ulm import Writer
from ase.units import Bohr, Ha
from gpaw.gpu import synchronize
from gpaw.gpu.mpi import CuPyMPI
from gpaw.mpi import MPIComm, serial_comm
from gpaw.new import cached_property, zips
from gpaw.new.brillouin import IBZ
from gpaw.new.c import GPU_AWARE_MPI
from gpaw.new.lcao.wave_functions import LCAOWaveFunctions
from gpaw.new.potential import Potential
from gpaw.new.pwfd.wave_functions import PWFDWaveFunctions
from gpaw.new.wave_functions import WaveFunctions
from gpaw.typing import Array1D, Array2D


def create_ibz_wave_functions(*,
                              ibz: IBZ,
                              nelectrons: float,
                              ncomponents: int,
                              create_wfs_func,
                              kpt_comm: MPIComm = serial_comm,
                              kpt_band_comm: MPIComm = serial_comm,
                              comm: MPIComm = serial_comm,
                              ) -> IBZWaveFunctions:
    """Collection of wave function objects for k-points in the IBZ."""
    rank_k = ibz.ranks(kpt_comm)
    mask_k = (rank_k == kpt_comm.rank)
    k_q = np.arange(len(ibz))[mask_k]

    nspins = ncomponents % 3

    wfs_qs: list[list[WaveFunctions]] = []
    for q, k in enumerate(k_q):
        wfs_s = []
        for spin in range(nspins):
            wfs = create_wfs_func(spin, q, k,
                                  ibz.kpt_kc[k], ibz.weight_k[k])
            wfs_s.append(wfs)
        wfs_qs.append(wfs_s)

    return IBZWaveFunctions(ibz,
                            nelectrons,
                            ncomponents,
                            wfs_qs,
                            kpt_comm,
                            kpt_band_comm,
                            comm)


class IBZWaveFunctions:
    def __init__(self,
                 ibz: IBZ,
                 nelectrons: float,
                 ncomponents: int,
                 wfs_qs: list[list[WaveFunctions]],
                 kpt_comm: MPIComm = serial_comm,
                 kpt_band_comm: MPIComm = serial_comm,
                 comm: MPIComm = serial_comm):
        """Collection of wave function objects for k-points in the IBZ."""
        self.ibz = ibz
        self.kpt_comm = kpt_comm
        self.kpt_band_comm = kpt_band_comm
        self.comm = comm
        self.nelectrons = nelectrons
        self.ncomponents = ncomponents
        self.collinear = (ncomponents != 4)
        self.spin_degeneracy = ncomponents % 2 + 1
        self.nspins = ncomponents % 3

        self.rank_k = ibz.ranks(kpt_comm)

        self.wfs_qs = wfs_qs

        self.q_k = {}  # IBZ-index to local index
        for wfs in self:
            self.q_k[wfs.k] = wfs.q

        self.band_comm = wfs.band_comm
        self.domain_comm = wfs.domain_comm
        self.dtype = wfs.dtype
        self.nbands = wfs.nbands

        self.fermi_levels: Array1D | None = None  # hartree

        self.energies: dict[str, float] = {}  # hartree

        self.xp = self.wfs_qs[0][0].xp
        if self.xp is not np:
            if not GPU_AWARE_MPI:
                self.kpt_comm = CuPyMPI(self.kpt_comm)  # type: ignore

    @cached_property
    def mode(self):
        wfs = self.wfs_qs[0][0]
        if isinstance(wfs, PWFDWaveFunctions):
            if hasattr(wfs.psit_nX.desc, 'ecut'):
                return 'pw'
            return 'fd'
        return 'lcao'

    def get_max_shape(self, global_shape: bool = False) -> tuple[int, ...]:
        """Find the largest wave function array shape.

        For a PW-calculation, this shape could depend on k-point.
        """
        if global_shape:
            shape = np.array(max(wfs.array_shape(global_shape=True)
                                 for wfs in self))
            self.kpt_comm.max(shape)
            return tuple(shape)
        return max(wfs.array_shape() for wfs in self)

    def __str__(self):
        shape = self.get_max_shape(global_shape=True)
        wfs = self.wfs_qs[0][0]
        nbytes = (len(self.ibz) *
                  self.nbands *
                  len(self.wfs_qs[0]) *
                  wfs.bytes_per_band)
        ncores = (self.kpt_comm.size *
                  self.domain_comm.size *
                  self.band_comm.size)
        return (f'{self.ibz.symmetries}\n'
                f'{self.ibz}\n'
                f'{wfs._short_string(shape)}\n'
                f'spin-components: {self.ncomponents}'
                '  # (' +
                ('' if self.collinear else 'non-') + 'collinear spins)\n'
                f'bands: {self.nbands}\n'
                f'valence electrons: {self.nelectrons}\n'
                f'spin-degeneracy: {self.spin_degeneracy}\n'
                f'dtype: {self.dtype}\n\n'
                'memory:\n'
                f'    wave functions: {nbytes:_}  # bytes '
                f' ({nbytes // ncores:_} per core)\n\n'
                'parallelization:\n'
                f'    kpt:    {self.kpt_comm.size}\n'
                f'    domain: {self.domain_comm.size}\n'
                f'    band:   {self.band_comm.size}\n')

    def __iter__(self) -> Generator[WaveFunctions, None, None]:
        for wfs_s in self.wfs_qs:
            yield from wfs_s

    def move(self, fracpos_ac, atomdist):
        self.ibz.symmetries.check_positions(fracpos_ac)
        self.energies.clear()
        for wfs in self:
            wfs.move(fracpos_ac, atomdist)

    def orthonormalize(self, work_array_nX: np.ndarray = None):
        for wfs in self:
            wfs.orthonormalize(work_array_nX)

    def calculate_occs(self, occ_calc, fixed_fermi_level=False):
        degeneracy = self.spin_degeneracy

        # u index is q and s combined
        occ_un, fermi_levels, e_entropy = occ_calc.calculate(
            nelectrons=self.nelectrons / degeneracy,
            eigenvalues=[wfs.eig_n * Ha for wfs in self],
            weights=[wfs.weight for wfs in self],
            fermi_levels_guess=(None
                                if self.fermi_levels is None else
                                self.fermi_levels * Ha))

        if not fixed_fermi_level:
            self.fermi_levels = np.array(fermi_levels) / Ha
        else:
            assert self.fermi_levels is not None

        for occ_n, wfs in zips(occ_un, self):
            wfs._occ_n = occ_n

        e_entropy *= degeneracy / Ha
        e_band = 0.0
        for wfs in self:
            e_band += wfs.occ_n @ wfs.eig_n * wfs.weight * degeneracy
        e_band = self.kpt_comm.sum(float(e_band))  # XXX CPU float?

        self.energies = {
            'band': e_band,
            'entropy': e_entropy,
            'extrapolation': e_entropy * occ_calc.extrapolate_factor}

    def add_to_density(self, nt_sR, D_asii) -> None:
        """Compute density and add to ``nt_sR`` and ``D_asii``."""
        for wfs in self:
            wfs.add_to_density(nt_sR, D_asii)

        if self.xp is not np:
            synchronize()

        # This should be done in a more efficient way!!!
        # Also: where do we want the density?
        self.kpt_comm.sum(nt_sR.data)
        self.kpt_comm.sum(D_asii.data)
        self.band_comm.sum(nt_sR.data)
        self.band_comm.sum(D_asii.data)

    def add_to_ked(self, taut_sR) -> None:
        for wfs in self:
            wfs.add_to_ked(taut_sR)
        if self.xp is not np:
            synchronize()
        self.kpt_comm.sum(taut_sR.data)
        self.band_comm.sum(taut_sR.data)

    def get_all_electron_wave_function(self,
                                       band,
                                       kpt=0,
                                       spin=0,
                                       grid_spacing=0.05,
                                       skip_paw_correction=False):
        wfs = self.get_wfs(kpt=kpt, spin=spin, n1=band, n2=band + 1)
        if wfs is None:
            return None
        assert isinstance(wfs, PWFDWaveFunctions)
        psit_X = wfs.psit_nX[0].to_pbc_grid()
        grid = psit_X.desc.uniform_grid_with_grid_spacing(grid_spacing)
        psi_r = psit_X.interpolate(grid=grid)

        if not skip_paw_correction:
            dphi_aj = wfs.setups.partial_wave_corrections()
            dphi_air = grid.atom_centered_functions(dphi_aj, wfs.fracpos_ac)
            dphi_air.add_to(psi_r, wfs.P_ani[:, 0])

        return psi_r

    def get_wfs(self,
                *,
                kpt: int = 0,
                spin: int = 0,
                n1=0,
                n2=0):
        rank = self.rank_k[kpt]
        if rank == self.kpt_comm.rank:
            wfs = self.wfs_qs[self.q_k[kpt]][spin]
            wfs2 = wfs.collect(n1, n2)
            if rank == 0:
                return wfs2
            if wfs2 is not None:
                wfs2.send(self.kpt_comm, 0)
            return
        master = (self.kpt_comm.rank == 0 and
                  self.domain_comm.rank == 0 and
                  self.band_comm.rank == 0)
        if master:
            return self.wfs_qs[0][0].receive(self.kpt_comm, rank)
        return None

    def get_eigs_and_occs(self, k=0, s=0):
        if self.domain_comm.rank == 0 and self.band_comm.rank == 0:
            rank = self.rank_k[k]
            if rank == self.kpt_comm.rank:
                wfs = self.wfs_qs[self.q_k[k]][s]
                if rank == 0:
                    return wfs._eig_n, wfs._occ_n
                self.kpt_comm.send(wfs._eig_n, 0)
                self.kpt_comm.send(wfs._occ_n, 0)
            elif self.kpt_comm.rank == 0:
                eig_n = np.empty(self.nbands)
                occ_n = np.empty(self.nbands)
                self.kpt_comm.receive(eig_n, rank)
                self.kpt_comm.receive(occ_n, rank)
                return eig_n, occ_n
        return np.zeros(0), np.zeros(0)

    def get_all_eigs_and_occs(self):
        nkpts = len(self.ibz)
        if self.comm.rank == 0:
            eig_skn = np.empty((self.nspins, nkpts, self.nbands))
            occ_skn = np.empty((self.nspins, nkpts, self.nbands))
        else:
            eig_skn = np.empty((0, 0, 0))
            occ_skn = np.empty((0, 0, 0))
        for k in range(nkpts):
            for s in range(self.nspins):
                eig_n, occ_n = self.get_eigs_and_occs(k, s)
                if self.comm.rank == 0:
                    eig_skn[s, k, :] = eig_n
                    occ_skn[s, k, :] = occ_n
        return eig_skn, occ_skn

    def forces(self, potential: Potential) -> Array2D:
        F_av = self.xp.zeros((potential.dH_asii.natoms, 3))
        for wfs in self:
            wfs.force_contribution(potential, F_av)
        if self.xp is not np:
            synchronize()
        self.kpt_comm.sum(F_av)
        return F_av

    def write(self,
              writer: Writer,
              skip_wfs: bool) -> None:
        """Write fermi-level(s), eigenvalues, occupation numbers, ...

        ... k-points, symmetry information, projections and possibly
        also the wave functions.
        """
        eig_skn, occ_skn = self.get_all_eigs_and_occs()
        assert self.fermi_levels is not None
        writer.write(fermi_levels=self.fermi_levels * Ha,
                     eigenvalues=eig_skn * Ha,
                     occupations=occ_skn)
        ibz = self.ibz
        writer.child('kpts').write(
            atommap=ibz.symmetries.a_sa,
            bz2ibz=ibz.bz2ibz_K,
            bzkpts=ibz.bz.kpt_Kc,
            ibzkpts=ibz.kpt_kc,
            rotations=ibz.symmetries.rotation_svv,
            translations=ibz.symmetries.translation_sc,
            weights=ibz.weight_k)

        nproj = self.wfs_qs[0][0].P_ani.layout.size

        spin_k_shape: tuple[int, ...]
        proj_shape: tuple[int, ...]

        if self.collinear:
            spin_k_shape = (self.ncomponents, len(ibz))
            proj_shape = (self.nbands, nproj)
        else:
            spin_k_shape = (len(ibz),)
            proj_shape = (self.nbands, 2, nproj)

        writer.add_array('projections', spin_k_shape + proj_shape, self.dtype)

        for spin in range(self.nspins):
            for k, rank in enumerate(self.rank_k):
                if rank == self.kpt_comm.rank:
                    wfs = self.wfs_qs[self.q_k[k]][spin]
                    P_ani = wfs.P_ani.to_cpu().gather()  # gather atoms
                    if P_ani is not None:
                        P_nI = P_ani.matrix.gather()  # gather bands
                        if P_nI is not None:
                            if rank == 0:
                                writer.fill(P_nI.data.reshape(proj_shape))
                            else:
                                self.kpt_comm.send(P_nI.data, 0)
                elif self.comm.rank == 0:
                    data = np.empty(proj_shape, self.dtype)
                    self.kpt_comm.receive(data, rank)
                    writer.fill(data)

        if skip_wfs:
            return

        xshape = self.get_max_shape(global_shape=True)
        shape = spin_k_shape + (self.nbands,) + xshape

        c = Bohr**-1.5
        if isinstance(wfs, LCAOWaveFunctions):
            c = 1

        for spin in range(self.nspins):
            for k, rank in enumerate(self.rank_k):
                if rank == self.kpt_comm.rank:
                    wfs = self.wfs_qs[self.q_k[k]][spin]
                    coef_nX = wfs.gather_wave_function_coefficients()
                    if coef_nX is not None:
                        if rank == 0:
                            if spin == 0 and k == 0:
                                writer.add_array('coefficients',
                                                 shape, dtype=coef_nX.dtype)
                            # For PW-mode, we may need to zero-pad the
                            # plane-wave coefficient up to the maximum
                            # for all k-points:
                            n = shape[-1] - coef_nX.shape[-1]
                            if n != 0:
                                if self.collinear:
                                    coef_nX = np.pad(coef_nX, ((0, 0), (0, n)))
                                else:
                                    coef_nX = np.pad(coef_nX,
                                                     ((0, 0), (0, 0), (0, n)))
                            writer.fill(coef_nX * c)
                        else:
                            self.kpt_comm.send(coef_nX, 0)
                elif self.comm.rank == 0:
                    assert coef_nX is not None
                    self.kpt_comm.receive(coef_nX, rank)
                    writer.fill(coef_nX * c)

    def write_summary(self, log):
        fl = self.fermi_levels * Ha
        if len(fl) == 1:
            log(f'\nFermi level: {fl[0]:.3f} eV')
        else:
            log(f'\nFermi levels: {fl[0]:.3f}, {fl[1]:.3f} eV')

        ibz = self.ibz

        eig_skn, occ_skn = self.get_all_eigs_and_occs()

        if self.comm.rank != 0:
            return

        eig_skn *= Ha

        D = self.spin_degeneracy
        nbands = eig_skn.shape[2]

        for k, (x, y, z) in enumerate(ibz.kpt_kc):
            if k == 3:
                log(f'(only showing first 3 out of {len(ibz)} k-points)')
                break

            log(f'\nkpt = [{x:.3f}, {y:.3f}, {z:.3f}], '
                f'weight = {ibz.weight_k[k]:.3f}:')

            if self.nspins == 1:
                skipping = False
                log(f'  Band      eig [eV]   occ [0-{D}]')
                eig_n = eig_skn[0, k]
                n0 = (eig_n < fl[0]).sum() - 0.5
                for n, (e, f) in enumerate(zips(eig_n, occ_skn[0, k])):
                    # First, last and +-8 bands window around Fermi level:
                    if n == 0 or abs(n - n0) < 8 or n == nbands - 1:
                        log(f'  {n:4} {e:13.3f}   {D * f:9.3f}')
                        skipping = False
                    else:
                        if not skipping:
                            log('   ...')
                            skipping = True
            else:
                log('  Band      eig [eV]   occ [0-1]'
                    '      eig [eV]   occ [0-1]')
                for n, (e1, f1, e2, f2) in enumerate(zips(eig_skn[0, k],
                                                          occ_skn[0, k],
                                                          eig_skn[1, k],
                                                          occ_skn[1, k])):
                    log(f'  {n:4} {e1:13.3f}   {f1:9.3f}'
                        f'    {e2:10.3f}   {f2:9.3f}')

        try:
            log()
            bandgap(eigenvalues=eig_skn,
                    efermi=fl[0],
                    output=log.fd,
                    kpts=ibz.kpt_kc)
        except ValueError:
            # Maybe we only have the occupied bands and no empty bands
            pass

    def make_sure_wfs_are_read_from_gpw_file(self):
        for wfs in self:
            psit_nX = getattr(wfs, 'psit_nX', None)
            if psit_nX is None:
                return
            if hasattr(psit_nX.data, 'fd'):  # fd=file-descriptor
                psit_nX.data = psit_nX.data[:]  # read

    def get_homo_lumo(self, spin: int = None) -> Array1D:
        """Return HOMO and LUMO eigenvalues."""
        if self.ncomponents == 1:
            N = 2
            assert spin != 1
            spin = 0
        elif self.ncomponents == 2:
            N = 2
            if spin is None:
                h0, l0 = self.get_homo_lumo(0)
                h1, l1 = self.get_homo_lumo(1)
                return np.array([max(h0, h1), min(l0, l1)])
        else:
            N = 1
            assert spin != 1
            spin = 0

        n = int(round(self.nelectrons)) // N
        assert N * n == self.nelectrons
        homo = self.kpt_comm.max(max(wfs_s[spin].eig_n[n - 1]
                                     for wfs_s in self.wfs_qs))
        lumo = self.kpt_comm.min(min(wfs_s[spin].eig_n[n]
                                     for wfs_s in self.wfs_qs))

        return np.array([homo, lumo])



import numpy as np
import pytest
from ase.build import molecule
from ase.units import Bohr, Hartree
from gpaw import GPAW
from gpaw.external import ConstantElectricField
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.lcaotddft.laser import GaussianPulse
from gpaw.mpi import world
from gpaw.tddft.units import as_to_au
from gpaw.test import equal

# Settings
dt = 20.0
N1 = 5
N = 5 + N1
kick_v = np.ones(3) * 1e-5


@pytest.mark.rttddft
def test_laser(in_tmp_dir):
    # Atoms
    atoms = molecule('Na2')
    atoms.center(vacuum=4.0)

    # Ground-state calculation
    calc = GPAW(nbands=2, h=0.4, setups=dict(Na='1'),
                basis='dzp', mode='lcao',
                convergence={'density': 1e-8},
                symmetry={'point_group': False},
                txt='gs.out')
    atoms.calc = calc
    _ = atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Time-propagation calculation
    td_calc = LCAOTDDFT('gs.gpw', txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick(kick_v)
    td_calc.propagate(dt, N)

    # Pulse
    direction = kick_v
    ext = ConstantElectricField(Hartree / Bohr, direction)
    pulse = GaussianPulse(1e-5, 0e3, 8.6, 0.5, 'sin')

    # Time-propagation calculation with pulse
    td_calc = LCAOTDDFT('gs.gpw', td_potential={'ext': ext, 'laser': pulse},
                        txt='tdpulse.out')
    DipoleMomentWriter(td_calc, 'dmpulse.dat')
    td_calc.propagate(dt, N1)
    td_calc.write('td.gpw', mode='all')
    # Restart
    td_calc = LCAOTDDFT('td.gpw', txt='tdpulse2.out')
    DipoleMomentWriter(td_calc, 'dmpulse.dat')
    td_calc.propagate(dt, N - N1)

    # Convoluted dipole moment
    world.barrier()
    time_t = np.arange(0, dt * (N + 0.1), dt) * as_to_au
    pulse_t = pulse.strength(time_t)
    np.savetxt('pulse.dat', np.stack((time_t, pulse_t)).T)
    dm_tv = np.delete(np.loadtxt('dm.dat')[:, 2:], 1, axis=0)
    dm_tv /= np.linalg.norm(kick_v)
    pulsedm_tv = np.delete(np.loadtxt('dmpulse.dat')[:, 2:], N1, axis=0)

    tol = 5e-6
    for v in range(3):
        pulsedmconv_t = np.convolve(
            dm_tv[:, v], pulse_t)[:(N + 1)] * dt * as_to_au
        np.savetxt('dmpulseconv%d.dat' % v, pulsedmconv_t)
        equal(pulsedm_tv[:, v], pulsedmconv_t, tol)



import numpy as np


class LinearSolver:
    def __init__(self,
                 method=None,
                 preconditioner=None,
                 tolerance=1e-10,
                 max_iter=10000):
        """Init method."""

        # self.solver = solvers[method]
        # self.pc = pcs[preconditioner]
        self.tolerance = tolerance
        self.max_iter = max_iter

    def solve(self, A, x_G, b_G):
        """Solve linear system A * x = b."""

        ##########################################
        # Simple steepest descent implementation #
        ##########################################

        # Residue: r = A*x - b
        r_G = np.zeros_like(x_G)

        A.dot(x_G, r_G)
        # A.project(r_G)
        r_G -= b_G

        for iter in range(self.max_iter):

            x_G -= 0.01 * r_G
            A.dot(x_G, r_G)
            # Try and project out the undersired components in the residue
            # A.project(r_G)
            r_G -= b_G
            error = A.norm(r_G)

            if error < self.tolerance:
                break

        return iter


'''
    def apply_preconditioner(self, x, b):
        """Solves preconditioner equation.

        Parameters
        ----------
        psi: List of coarse grids
            the known wavefunctions
        psin: List of coarse grids
            the result

        """

        #self.timer.start('Solve Sternheimer preconditioner')
        if self.preconditioner is not None:
            #self.preconditioner.apply(self.kpt, psi, psin)
            pass
        else:
            psin[:] = psi
        #self.timer.stop('Solve Sternheimer preconditioner')
'''



"""Brillouin-zone sampling."""
from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from ase.dft.kpoints import monkhorst_pack
from gpaw.mpi import MPIComm
from gpaw.typing import Array1D, ArrayLike2D
if TYPE_CHECKING:
    from gpaw.new.symmetry import Symmetries


class BZPoints:
    def __init__(self, points: ArrayLike2D):
        self.kpt_Kc = np.array(points)
        assert self.kpt_Kc.ndim == 2
        assert self.kpt_Kc.shape[1] == 3
        self.gamma_only = len(self.kpt_Kc) == 1 and not self.kpt_Kc.any()

    def __len__(self):
        """Number of k-points in the BZ."""
        return len(self.kpt_Kc)

    def __repr__(self):
        if self.gamma_only:
            return 'BZPoints([<gamma only>])'
        return f'BZPoints([<{len(self)} points>])'


class MonkhorstPackKPoints(BZPoints):
    def __init__(self, size, shift=(0, 0, 0)):
        self.size_c = size
        self.shift_c = np.array(shift)
        super().__init__(monkhorst_pack(size) + shift)

    def __repr__(self):
        return f'MonkhorstPackKPoints({self.size_c}, shift={self.shift_c})'

    def __str__(self):
        a, b, c = self.size_c
        l, m, n = self.shift_c
        return (f'monkhorst-pack size: [{a}, {b}, {c}]\n'
                f'monkhorst-pack shift: [{l}, {m}, {n}]\n')


class IBZ:
    def __init__(self,
                 symmetries: Symmetries,
                 bz: BZPoints,
                 ibz2bz, bz2ibz, weights):
        self.symmetries = symmetries
        self.bz = bz
        self.weight_k = weights
        self.kpt_kc = bz.kpt_Kc[ibz2bz]
        self.ibz2bz_k = ibz2bz
        self.bz2ibz_K = bz2ibz

        # self.bz2bz_Ks = []  # later ...

    def __len__(self):
        """Number of k-points in the IBZ."""
        return len(self.kpt_kc)

    def __repr__(self):
        return (f'IBZ(<points: {len(self)}, '
                f'symmetries: {len(self.symmetries)}>)')

    def __str__(self):
        N = len(self)
        txt = ('bz sampling:\n'
               f'  number of bz points: {len(self.bz)}\n'
               f'  number of ibz points: {N}\n')

        if isinstance(self.bz, MonkhorstPackKPoints):
            txt += '  ' + str(self.bz).replace('\n', '\n  ', 1)

        txt += '  points and weights: [\n'
        k = 0
        while k < N:
            if k == 10:
                if N > 10:
                    txt += '    # ...\n'
                k = N - 1
            a, b, c = self.kpt_kc[k]
            w = self.weight_k[k]
            t = ',' if k < N - 1 else ']'
            txt += (f'    [[{a:12.8f}, {b:12.8f}, {c:12.8f}], '
                    f'{w:.8f}]{t}  # {k}\n')
            k += 1
        return txt

    def ranks(self, comm: MPIComm) -> Array1D:
        """Distribute k-points over MPI-communicator."""
        return ranks(comm.size, len(self))


def ranks(N, K) -> Array1D:
    """Distribute k-points over MPI-communicator.

    >>> ranks(4, 6)
    array([0, 1, 2, 2, 3, 3])
    """
    n, x = divmod(K, N)
    rnks = np.empty(K, int)
    r = N - x
    for k in range(r * n):
        rnks[k] = k // n
    for k in range(r * n, K):
        rnks[k] = (k - r * n) // (n + 1) + r
    return rnks



"""New ground-state DFT code."""
from __future__ import annotations
from collections import defaultdict
from contextlib import contextmanager
from time import time
from typing import Iterable, TYPE_CHECKING
if TYPE_CHECKING:
    from gpaw.core import UGArray


def prod(iterable: Iterable[int]) -> int:
    """Simple int product.

    >>> prod([])
    1
    >>> prod([2, 3])
    6
    """
    result = 1
    for x in iterable:
        result *= x
    return result


def cached_property(method):
    """Quick'n'dirty implementation of cached_property coming in Python 3.8."""
    name = f'__{method.__name__}'

    def new_method(self):
        if not hasattr(self, name):
            setattr(self, name, method(self))
        return getattr(self, name)

    return property(new_method)


def zips(*iterables, strict=True):
    """From PEP 618."""
    if not iterables:
        return
    iterators = tuple(iter(iterable) for iterable in iterables)
    try:
        while True:
            items = []
            for iterator in iterators:
                items.append(next(iterator))
            yield tuple(items)
    except StopIteration:
        pass
    if not strict:
        return
    if items:
        i = len(items)
        plural = " " if i == 1 else "s 1-"
        msg = f"zips() argument {i+1} is shorter than argument{plural}{i}"
        raise ValueError(msg)
    sentinel = object()
    for i, iterator in enumerate(iterators[1:], 1):
        if next(iterator, sentinel) is not sentinel:
            plural = " " if i == 1 else "s 1-"
            msg = f"zips() argument {i+1} is longer than argument{plural}{i}"
            raise ValueError(msg)


def spinsum(a_sX: UGArray, mean: bool = False) -> UGArray:
    if a_sX.dims[0] == 2:
        a_X = a_sX.desc.empty()
        a_sX.data[:2].sum(axis=0, out=a_X.data)
        if mean:
            a_X.data *= 0.5
        return a_X
    return a_sX[0]


class Timer:
    def __init__(self):
        self.times = defaultdict(float)
        self.times['Total'] = -time()

    @contextmanager
    def __call__(self, name):
        t1 = time()
        try:
            yield
        finally:
            t2 = time()
            self.times[name] += t2 - t1

    def write(self, log):
        self.times['Total'] += time()
        total = self.times['Total']
        log('\ntiming:  # [seconds]')
        n = max(len(name) for name in self.times) + 2
        w = len(f'{total:.3f}')
        N = 71 - n - w
        for name, t in self.times.items():
            m = int(round(2 * N * t / total))
            bar = '━' * (m // 2) + '╸' * (m % 2)
            log(f'  {name + ":":{n}}{t:{w}.3f}  # {bar}')



import pytest
from ase.build import bulk
from gpaw import GPAW, FermiDirac
from gpaw.mpi import serial_comm
from gpaw.test import equal
from gpaw.xc.rpa import RPACorrelation
from gpaw.xc.fxc import FXCCorrelation


@pytest.mark.rpa
@pytest.mark.response
def test_rpa_rpa_energy_Ni(in_tmp_dir):
    Ni = bulk('Ni', 'fcc')
    Ni.set_initial_magnetic_moments([0.7])

    calc = GPAW(mode='pw',
                kpts=(3, 3, 3),
                occupations=FermiDirac(0.001),
                setups={'Ni': '10'},
                communicator=serial_comm)
    Ni.calc = calc
    Ni.get_potential_energy()
    calc.diagonalize_full_hamiltonian(nbands=50)

    rpa = RPACorrelation(calc, nfrequencies=8, skip_gamma=True, ecut=[50])
    E_rpa = rpa.calculate()

    fxc = FXCCorrelation(calc, nlambda=16, nfrequencies=8, skip_gamma=True,
                         ecut=[50])
    E_fxc = fxc.calculate()

    equal(E_rpa, -7.826, 0.01)
    equal(E_fxc, -7.826, 0.01)



import numpy as np
from ase import Atoms
from ase.units import Bohr

from gpaw.band_descriptor import BandDescriptor
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.new import cached_property, prod, zips
from gpaw.new.calculation import DFTCalculation
from gpaw.new.pwfd.wave_functions import PWFDWaveFunctions
from gpaw.projections import Projections
from gpaw.pw.descriptor import PWDescriptor
from gpaw.utilities import pack
from gpaw.wavefunctions.arrays import PlaneWaveExpansionWaveFunctions


class FakeWFS:
    def __init__(self, calculation: DFTCalculation, atoms: Atoms):
        self.setups = calculation.setups
        self.state = calculation.state
        ibzwfs = self.state.ibzwfs
        self.kd = KPointDescriptor(ibzwfs.ibz.bz.kpt_Kc,
                                   ibzwfs.nspins)
        self.kd.set_symmetry(atoms,
                             ibzwfs.ibz.symmetries.symmetry)
        self.kd.set_communicator(ibzwfs.kpt_comm)
        self.bd = BandDescriptor(ibzwfs.nbands, ibzwfs.band_comm)
        grid = self.state.density.nt_sR.desc
        self.gd = grid._gd
        self.atom_partition = calculation._atom_partition
        self.setups.set_symmetry(ibzwfs.ibz.symmetries.symmetry)
        self.occupations = calculation.scf_loop.occ_calc.occ
        self.nvalence = int(round(ibzwfs.nelectrons))
        assert self.nvalence == ibzwfs.nelectrons
        self.world = calculation.comm
        if ibzwfs.fermi_levels is not None:
            self.fermi_levels = ibzwfs.fermi_levels
            if len(self.fermi_levels) == 1:
                self.fermi_level = self.fermi_levels[0]
        self.nspins = ibzwfs.nspins
        self.dtype = ibzwfs.dtype
        wfs = ibzwfs.wfs_qs[0][0]
        self.pd = None
        if isinstance(wfs, PWFDWaveFunctions):
            if hasattr(wfs.psit_nX.desc, 'ecut'):
                self.mode = 'pw'
                self.pd = PWDescriptor(wfs.psit_nX.desc.ecut,
                                       self.gd, self.dtype, self.kd)
                self.pwgrid = grid.new(dtype=self.dtype)
            else:
                self.mode = 'fd'
        else:
            self.mode = 'lcao'

    def _get_wave_function_array(self, u, n, realspace):
        psit_X = self.kpt_u[u].wfs.psit_nX[n]
        if hasattr(psit_X, 'ifft'):
            return psit_X.ifft(grid=self.pwgrid).data
        return psit_X.data

    def get_wave_function_array(self, n, k, s, realspace=True, periodic=False):
        assert realspace
        psit_X = self.kpt_qs[k][s].wfs.psit_nX[n]
        if self.mode == 'pw':
            psit_R = psit_X.ifft(grid=self.pwgrid, periodic=periodic)
        else:
            psit_R = psit_X
            if periodic:
                psit_R.multiply_by_eikr(-psit_R.desc.kpt_c)
        return psit_R.data

    def collect_projections(self, k, s):
        return self.kpt_qs[k][s].projections.collect()

    def collect_eigenvalues(self, k, s):
        return self.state.ibzwfs.wfs_qs[k][s].eig_n.copy()

    @cached_property
    def kpt_u(self):
        return [kpt
                for kpt_s in self.kpt_qs
                for kpt in kpt_s]

    @cached_property
    def kpt_qs(self):
        ngpts = prod(self.gd.N_c)
        return [[KPT(wfs, self.atom_partition, ngpts, self.pd)
                 for wfs in wfs_s]
                for wfs_s in self.state.ibzwfs.wfs_qs]


class KPT:
    def __init__(self, wfs, atom_partition, ngpts, pd):
        self.ngpts = ngpts
        self.wfs = wfs
        self.pd = pd

        I1 = 0
        nproj_a = []
        for a, shape in enumerate(wfs.P_ani.layout.shape_a):
            I2 = I1 + prod(shape)
            nproj_a.append(I2 - I1)
            I1 = I2

        self.projections = Projections(
            wfs.nbands,
            nproj_a,
            atom_partition,
            wfs.P_ani.comm,
            wfs.ncomponents < 4,
            wfs.spin,
            data=wfs.P_ani.data)
        self.eps_n = wfs.eig_n
        self.s = wfs.spin if wfs.ncomponents < 4 else None
        self.k = wfs.k
        self.q = wfs.q
        self.weight = wfs.spin_degeneracy * wfs.weight
        self.f_n = wfs.occ_n * self.weight
        self.P_ani = wfs.P_ani
        if isinstance(wfs, PWFDWaveFunctions):
            self.psit_nX = wfs.psit_nX

    @property
    def psit_nG(self):
        a_nG = self.psit_nX.data
        if a_nG.ndim == 4:
            return a_nG
        return a_nG * self.ngpts

    @cached_property
    def psit(self):
        return PlaneWaveExpansionWaveFunctions(
            self.wfs.nbands, self.pd, self.wfs.dtype,
            self.psit_nX.data * self.ngpts,
            kpt=self.q,
            dist=(),  # self.bd.comm, self.bd.comm.size),
            spin=self.s,
            collinear=self.wfs.ncomponents != 4)


class FakeDensity:
    def __init__(self, calculation: DFTCalculation):
        self.setups = calculation.setups
        self.state = calculation.state
        self.D_asii = self.state.density.D_asii
        self.atom_partition = calculation._atom_partition
        self.interpolate = calculation.pot_calc._interpolate_density
        self.nt_sR = self.state.density.nt_sR
        self.nt_sG = self.nt_sR.data
        self.gd = self.nt_sR.desc._gd
        self.finegd = calculation.pot_calc.fine_grid._gd
        self._densities = calculation.densities()
        self.ncomponents = len(self.nt_sG)
        self.nspins = self.ncomponents % 3

    @cached_property
    def D_asp(self):
        D_asp = self.setups.empty_atomic_matrix(self.ncomponents,
                                                self.atom_partition)
        D_asp.update({a: np.array([pack(D_ii) for D_ii in D_sii.real])
                      for a, D_sii in self.D_asii.items()})
        return D_asp

    @cached_property
    def nt_sg(self):
        return self.interpolate(self.nt_sR)[0].data

    def interpolate_pseudo_density(self):
        pass

    def get_all_electron_density(self, *, atoms, gridrefinement):
        n_sr = self._densities.all_electron_densities(
            grid_refinement=gridrefinement).scaled(1 / Bohr, Bohr**3)
        return n_sr.data, n_sr.desc._gd


class FakeHamiltonian:
    def __init__(self, calculation: DFTCalculation):
        self.pot_calc = calculation.pot_calc
        self.finegd = self.pot_calc.fine_grid._gd
        self.grid = calculation.state.potential.vt_sR.desc
        self.e_total_free = calculation.results.get('free_energy')
        self.e_xc = calculation.state.potential.energies['xc']

    def restrict_and_collect(self, vxct_sg):
        fine_grid = self.pot_calc.fine_grid
        vxct_sr = fine_grid.empty(len(vxct_sg))
        vxct_sr.data[:] = vxct_sg
        vxct_sR = self.grid.empty(len(vxct_sg))
        for vxct_r, vxct_R in zips(vxct_sr, vxct_sR):
            self.pot_calc.restrict(vxct_r, vxct_R)
        return vxct_sR.data

    @property
    def xc(self):
        return self.pot_calc.xc.xc



import os
import sys
from optparse import OptionParser


def build_parser():
    usage = '%prog [OPTION] [BASIS]...'
    parser = OptionParser(usage=usage, version='%prog 1.0')
    parser.add_option('-f', '--files', action='store_true',
                      dest='actual_filenames',
                      help='Read from specified filenames rather than '
                      'searching GPAW setup directories')
    parser.add_option('-s', '--save-figs', action='store_true', dest='save',
                      help='Save figures to disk rather than showing plots')
    parser.add_option('-l', '--literal', action='store_true',
                      help='Do not pre-multiply wave functions by r in plots')
    parser.add_option('-n', '--normalize', action='store_true',
                      help='Plot normalized wave functions')
    parser.add_option('-x', '--ext', default='png',
                      help='Image format [default: %default]')
    return parser


def main(argv=sys.argv[1:]):
    parser = build_parser()
    opts, files = parser.parse_args(argv)

    import matplotlib.pyplot as plt
    from gpaw.basis_data import Basis, BasisPlotter

    plotter = BasisPlotter(premultiply=not opts.literal,
                           normalize=opts.normalize,
                           show=False,
                           save=opts.save,
                           ext=opts.ext)

    for path in files:
        dir, filename = os.path.split(path)
        splitfilename = filename.split('.')
        symbol = splitfilename[0]
        name = '.'.join(splitfilename[1:-1])
        if opts.actual_filenames:
            basis = Basis(symbol, name, False)
            basis.read_xml(path)
        else:  # search GPAW setup dirs
            basis = Basis(symbol, name)
        plotter.plot(basis)

    if not opts.save:
        plt.show()



import warnings

from ase.parallel import parprint
from ase.units import Bohr
import numpy as np
from scipy.special import erf

from gpaw.solvation.poisson import (WeightedFDPoissonSolver,
                                    ADM12PoissonSolver,
                                    PolarizationPoissonSolver)
from gpaw.solvation.dielectric import Dielectric
from gpaw.grid_descriptor import GridDescriptor
from gpaw.utilities.gauss import Gaussian
from gpaw.fd_operators import Gradient
from gpaw.test import equal

nn = 3
accuracy = 2e-10


def gradient(gd, x, nn):
    out = gd.empty(3)
    for i in (0, 1, 2):
        Gradient(gd, i, 1.0, nn).apply(x, out[i])
    return out


def make_gd(h, box, pbc):
    diag = np.array([box] * 3)
    cell = np.diag(diag)
    grid_shape = tuple((diag / h * 2).astype(int))
    return GridDescriptor(grid_shape, cell / Bohr, pbc)


class MockDielectric(Dielectric):
    def __init__(self, epsinf, nn):
        Dielectric.__init__(self, epsinf)
        self.nn = nn

    def update(self, cavity):
        grad_eps_vg = gradient(self.gd, self.eps_gradeps[0] - self.epsinf,
                               self.nn)
        for i in (0, 1, 2):
            self.eps_gradeps[1 + i][...] = grad_eps_vg[i]


def test_solvation_poisson():
    box = 12.
    gd = make_gd(h=.4, box=box, pbc=False)

    def solve(ps, eps, rho):
        phi = gd.zeros()
        dielectric = MockDielectric(epsinf=eps.max(), nn=nn)
        dielectric.set_grid_descriptor(gd)
        dielectric.allocate()
        dielectric.eps_gradeps[0][...] = eps
        dielectric.update(None)
        with warnings.catch_warnings():
            # ignore production code warning for alternative psolvers
            warnings.simplefilter("ignore")
            solver = ps(nn=nn, relax='J', eps=accuracy)
        solver.set_dielectric(dielectric)
        solver.set_grid_descriptor(gd)
        solver.solve(phi, rho)
        return phi

    psolvers = (WeightedFDPoissonSolver,
                ADM12PoissonSolver,
                PolarizationPoissonSolver)

    # test neutral system with constant permittivity
    parprint('neutral, constant permittivity')
    epsinf = 80.
    eps = gd.zeros()
    eps.fill(epsinf)
    qs = (-1., 1.)
    shifts = (-1., 1.)
    rho = gd.zeros()
    phi_expected = gd.zeros()
    for q, shift in zip(qs, shifts):
        gauss_norm = q / np.sqrt(4 * np.pi)
        gauss = Gaussian(gd, center=(box / 2. + shift) * np.ones(3) / Bohr)
        rho += gauss_norm * gauss.get_gauss(0)
        phi_expected += gauss_norm * gauss.get_gauss_pot(0) / epsinf

    for ps in psolvers:
        phi = solve(ps, eps, rho)
        parprint(ps, np.abs(phi - phi_expected).max())
        equal(phi, phi_expected, 1e-3)

    # test charged system with constant permittivity
    parprint('charged, constant permittivity')
    epsinf = 80.
    eps = gd.zeros()
    eps.fill(epsinf)
    q = -2.
    gauss_norm = q / np.sqrt(4 * np.pi)
    gauss = Gaussian(gd, center=(box / 2. + 1.) * np.ones(3) / Bohr)
    rho_gauss = gauss_norm * gauss.get_gauss(0)
    phi_gauss = gauss_norm * gauss.get_gauss_pot(0)
    phi_expected = phi_gauss / epsinf

    for ps in psolvers:
        phi = solve(ps, eps, rho_gauss)
        parprint(ps, np.abs(phi - phi_expected).max())
        equal(phi, phi_expected, 1e-3)

    # test non-constant permittivity
    msgs = ('neutral, non-constant permittivity',
            'charged, non-constant permittivity')
    qss = ((-1., 1.), (2., ))
    shiftss = ((-.4, .4), (-1., ))
    epsshifts = (.0, -1.)

    for msg, qs, shifts, epsshift in zip(msgs, qss, shiftss, epsshifts):
        parprint(msg)
        epsinf = 80.
        gauss = Gaussian(gd, center=(box / 2. + epsshift) * np.ones(3) / Bohr)
        eps = gauss.get_gauss(0)
        eps = epsinf - eps / eps.max() * (epsinf - 1.)

        rho = gd.zeros()
        phi_expected = gd.zeros()
        grad_eps = gradient(gd, eps - epsinf, nn)

        for q, shift in zip(qs, shifts):
            gauss = Gaussian(gd, center=(box / 2. + shift) * np.ones(3) / Bohr)
            phi_tmp = gauss.get_gauss_pot(0)
            xyz = gauss.xyz
            fac = 2. * np.sqrt(gauss.a) * np.exp(-gauss.a * gauss.r2)
            fac /= np.sqrt(np.pi) * gauss.r2
            fac -= erf(np.sqrt(gauss.a) * gauss.r) / (gauss.r2 * gauss.r)
            fac *= 2.0 * 1.7724538509055159
            grad_phi = fac * xyz
            laplace_phi = -4. * np.pi * gauss.get_gauss(0)
            rho_tmp = -1. / (4. * np.pi) * (
                (grad_eps * grad_phi).sum(0) + eps * laplace_phi)
            norm = gd.integrate(rho_tmp)
            rho_tmp /= norm * q
            phi_tmp /= norm * q
            rho += rho_tmp
            phi_expected += phi_tmp

        for ps in psolvers:
            phi = solve(ps, eps, rho)
            parprint(ps, np.abs(phi - phi_expected).max())
            equal(phi, phi_expected, 1e-3)



import numbers
from abc import ABC, abstractmethod
from math import factorial as fac
from math import pi
from typing import Tuple

import numpy as np
from gpaw.sphere.integrate import integrate_radial_grid
from gpaw.spline import Spline
from gpaw.typing import Array1D
from gpaw.utilities import divrl, hartree
from scipy.interpolate import make_interp_spline, splder


def radial_grid_descriptor(eq: str, **kwargs) -> 'RadialGridDescriptor':
    if eq == 'r=d*i':
        assert int(kwargs['istart']) == 0
        return EquidistantRadialGridDescriptor(float(kwargs['d']),
                                               int(kwargs['iend']) + 1)
    if eq == 'r=a*i/(n-i)':
        beta = float(kwargs['a'])
        ng = int(kwargs['n'])
        return AERadialGridDescriptor(beta / ng, 1.0 / ng, ng)
    if eq == 'r=a*i/(1-b*i)':
        a = float(kwargs['a'])
        b = float(kwargs['b'])
        N = int(kwargs['n'])
        return AERadialGridDescriptor(a, b, N)
    if eq == 'r=a*(exp(d*i)-1)':
        a = float(kwargs['a'])
        d = float(kwargs['d'])
        N = int(kwargs['iend']) + 1
        return AbinitRadialGridDescriptor(a, d, N)
    raise ValueError('Unknown grid: ' + eq)


def fsbt(l, f_g, r_g, G_k):
    """Fast spherical Bessel transform.

    Returns::

          oo
         / 2
         |r dr j (Gr) f(r),
         /      l
          0

    using l+1 fft's."""

    N = (len(G_k) - 1) * 2
    f_k = 0.0
    F_g = f_g * r_g
    for n in range(l + 1):
        f_k += (r_g[1] * (1j)**(l + 1 - n) *
                fac(l + n) / fac(l - n) / fac(n) / 2**n *
                np.fft.rfft(F_g, N)).real * G_k**(l - n)
        F_g[1:] /= r_g[1:]

    f_k[1:] /= G_k[1:]**(l + 1)
    if l == 0:
        f_k[0] = np.dot(r_g, f_g * r_g) * r_g[1]
    return f_k


class RadialGridDescriptor(ABC):
    ndim = 1  # dimension of ndarrays

    def __init__(self, r_g: np.ndarray, dr_g, default_spline_points=25):
        """Grid descriptor for radial grid."""
        self.r_g = r_g
        self.dr_g = dr_g
        self.N = len(r_g)
        self.dv_g = 4 * pi * r_g**2 * dr_g
        self.default_spline_points = default_spline_points

    def __len__(self):
        return self.N

    @classmethod
    def new(cls, name, *args, **kwargs):
        if name == 'equidistant':
            return EquidistantRadialGridDescriptor(*args, **kwargs)
        raise ValueError(f'Unknown grid-type: {name}')

    def zeros(self, x=()):
        a_xg = self.empty(x)
        a_xg[:] = 0
        return a_xg

    def empty(self, x=()):
        if isinstance(x, int):
            x = (x,)
        return np.zeros(x + (self.N,))

    def integrate(self, a_xg, n=0):
        assert n >= -2
        return np.dot(a_xg[..., 1:],
                      (self.r_g**(2 + n) * self.dr_g)[1:]) * (4 * pi)

    def integrate_trapz(self, a_xg, rcut=None):
        return integrate_radial_grid(a_xg, self.r_g, rcut=rcut)

    def yukawa(self, n_g, l=0, gamma=1e-6):
        r"""Calculates the radial grid yukawa integral.

        The the integral kernel for the Yukawa interaction:

                    \    _   _
              exp(- /\ | r - r' |)
              ----------------------
                      _   _
                    | r - r' |

           is defined as

            __    __            \  r              \  r    * ^     ^
          \     4 ||  I_(l+0.5)(/\  <) K_(l+0.5) (/\  >) Y (r)  Y(r')
           )          --------------------------          lm     lm
          / __            (rr')^0.5
            lm

         where I and K are the modified Bessel functions of the first
         and second kind (K is also known as Macdonald function).
         r = min (r, r')     r = max(r, r')
          <                   >
         We now calculate the integral:


                  ^    / _           ^
         v (r) Y (r) = |dr' n(r') Y (r')
          l     lm     /     l     lm

        with the Yukawa kernel mentioned above.

        And the output array is 'vr' as it is
        within the Hartree / radial Poisson solver.
        """

        from scipy.special import iv, kv
        vr_g = self.zeros()
        nrdr_g = n_g * self.r_g**1.5 * self.dr_g
        p = 0
        q = 0
        k_rgamma = kv(l + 0.5, self.r_g * gamma)      # K(>)
        i_rgamma = iv(l + 0.5, self.r_g * gamma)      # I(<)
        k_rgamma[0] = kv(l + 0.5, self.r_g[1] * gamma * 1e-5)
        # We have two integrals: one for r< and one for r>
        # This loop-technique helps calculate them in once
        for g_ind in range(len(nrdr_g) - 1, -1, -1):
            dp = k_rgamma[g_ind] * nrdr_g[g_ind]  # r' is r>
            dq = i_rgamma[g_ind] * nrdr_g[g_ind]  # r' is r<
            vr_g[g_ind] = (p + 0.5 * dp) * i_rgamma[g_ind] - \
                          (q + 0.5 * dq) * k_rgamma[g_ind]
            p += dp
            q += dq
        vr_g[:] += q * k_rgamma[:]
        vr_g *= 4 * pi
        vr_g[:] *= self.r_g[:]**0.5
        return vr_g

    def derivative_spline(self, n_g, dndr_g=None, *, degree=5):
        """Spline-based derivative of radial function."""
        if dndr_g is None:
            dndr_g = self.empty()

        # Boundary conditions
        assert degree in [3, 5]
        if degree == 5:
            bc_type = ([(2, 0.0), (4, 0.0)], [(2, 0.0), (4, 0.0)])
        elif degree == 3:
            bc_type = ([(2, 0.0)], [(2, 0.0)])

        s = make_interp_spline(self.r_g, n_g, k=degree, bc_type=bc_type)
        s = splder(s)
        dndr_g[:] = s(self.r_g)
        return dndr_g

    def derivative(self, n_g, dndr_g=None):
        """Finite-difference derivative of radial function."""
        if dndr_g is None:
            dndr_g = self.empty()
        dndr_g[0] = n_g[1] - n_g[0]
        dndr_g[1:-1] = 0.5 * (n_g[2:] - n_g[:-2])
        dndr_g[-1] = n_g[-1] - n_g[-2]
        dndr_g /= self.dr_g
        return dndr_g

    def derivative2(self, a_g, b_g):
        """Finite-difference derivative of radial function.

        For an infinitely dense grid, this method would be identical
        to the `derivative` method."""

        c_g = a_g / self.dr_g
        b_g[0] = 0.5 * c_g[1] + c_g[0]
        b_g[1] = 0.5 * c_g[2] - c_g[0]
        b_g[1:-1] = 0.5 * (c_g[2:] - c_g[:-2])
        b_g[-2] = c_g[-1] - 0.5 * c_g[-3]
        b_g[-1] = -c_g[-1] - 0.5 * c_g[-2]

    def laplace(self, n_g, d2ndr2_g=None):
        """Laplace of radial function."""
        if d2ndr2_g is None:
            d2ndr2_g = self.empty()
        dndg_g = 0.5 * (n_g[2:] - n_g[:-2])
        d2ndg2_g = n_g[2:] - 2 * n_g[1:-1] + n_g[:-2]
        d2ndr2_g[1:-1] = (d2ndg2_g / self.dr_g[1:-1]**2 +
                          dndg_g * (self.d2gdr2()[1:-1] +
                                    2 / self.r_g[1:-1] / self.dr_g[1:-1]))
        d2ndr2_g[0] = d2ndr2_g[1]
        d2ndr2_g[-1] = d2ndr2_g[-2]
        return d2ndr2_g

    def T(self, u_g, l):
        dudg_g = 0.5 * (u_g[2:] - u_g[:-2])
        d2udg2_g = u_g[2:] - 2 * u_g[1:-1] + u_g[:-2]
        Tu_g = self.empty()
        Tu_g[1:-1] = -0.5 * (d2udg2_g / self.dr_g[1:-1]**2 +
                             dudg_g * self.d2gdr2()[1:-1])
        Tu_g[-1] = Tu_g[-2]
        Tu_g[1:] += 0.5 * l * (l + 1) * u_g[1:] / self.r_g[1:]**2
        Tu_g[0] = Tu_g[1]
        return Tu_g

    def interpolate(self, f_g, r_x):
        from scipy.interpolate import InterpolatedUnivariateSpline
        return InterpolatedUnivariateSpline(self.r_g, f_g)(r_x)

    def fft(self, fr_g, l=0, N=None):
        """Fourier transform.

        Returns G and f(G) arrays::

                                          _ _
               l    ^    / _         ^   iG.r
          f(G)i Y  (G) = |dr f(r)Y  (r) e    .
                 lm      /        lm
        """

        if N is None:
            N = 2**13

        assert N % 2 == 0

        r_x = np.linspace(0, self.r_g[-1], N)
        f_x = self.interpolate(fr_g, r_x)
        f_x[1:] /= r_x[1:]
        f_x[0] = f_x[1]
        G_k = np.linspace(0, pi / r_x[1], N // 2 + 1)
        f_k = 4 * pi * fsbt(l, f_x, r_x, G_k)
        return G_k, f_k

    def filter(self, f_g, rcut, Gcut, l=0):
        Rcut = 100.0
        N = 1024 * 8
        r_x = np.linspace(0, Rcut, N, endpoint=False)
        h = Rcut / N

        alpha = 4.0 / rcut**2
        mcut = np.exp(-alpha * rcut**2)
        r2_x = r_x**2
        m_x = np.exp(-alpha * r2_x)
        for n in range(2):
            m_x -= (alpha * (rcut**2 - r2_x))**n * (mcut / fac(n))
        xcut = int(np.ceil(rcut / r_x[1]))
        m_x[xcut:] = 0.0

        G_k = np.linspace(0, pi / h, N // 2 + 1)

        # Zeropad the function to same length as coordinates:
        fpad_g = np.zeros(len(self.r_g))
        fpad_g[:len(f_g)] = f_g
        f_g = fpad_g

        from scipy.interpolate import InterpolatedUnivariateSpline
        if l < 2:
            f_x = InterpolatedUnivariateSpline(self.r_g, f_g)(r_x)
        else:
            a_g = f_g.copy()
            a_g[1:] /= self.r_g[1:]**(l - 1)
            f_x = InterpolatedUnivariateSpline(
                self.r_g, a_g)(r_x) * r_x**(l - 1)

        f_x[:xcut] /= m_x[:xcut]
        f_k = fsbt(l, f_x, r_x, G_k)
        kcut = int(Gcut / G_k[1])
        f_k[kcut:] = 0.0
        ff_x = fsbt(l, f_k, G_k, r_x[:N // 2 + 1]) / pi * 2
        ff_x *= m_x[:N // 2 + 1]

        if l < 2:
            f_g = InterpolatedUnivariateSpline(
                r_x[:xcut + 1], ff_x[:xcut + 1])(self.r_g)
        else:
            ff_x[1:xcut + 1] /= r_x[1:xcut + 1]**(l - 1)
            f_g = InterpolatedUnivariateSpline(
                r_x[:xcut + 1], ff_x[:xcut + 1])(self.r_g) * self.r_g**(l - 1)
        f_g[self.ceil(rcut):] = 0.0

        return f_g

    def poisson(self, n_g, l=0):
        vr_g = self.zeros()
        nrdr_g = n_g * self.r_g * self.dr_g
        hartree(l, nrdr_g, self.r_g, vr_g)
        return vr_g

    def pseudize(self, a_g, gc, l=0, points=4):
        """Construct smooth continuation of a_g for g<gc.

        Returns (b_g, c_p[P-1]) such that b_g=a_g for g >= gc and::

                P-1      2(P-1-p)+l
            b = Sum c_p r
             g  p=0      g

        for g < gc+P.
        """
        assert isinstance(gc, numbers.Integral) and gc > 10, gc

        r_g = self.r_g
        i = np.arange(gc, gc + points)
        r_i = r_g[i]
        c_p = np.polyfit(r_i**2, a_g[i] / r_i**l, points - 1)
        b_g = a_g.copy()
        b_g[:gc] = np.polyval(c_p, r_g[:gc]**2) * r_g[:gc]**l
        return b_g, c_p[-1]

    def cut(self, a_g, rcut):
        gcut = self.floor(rcut)
        r0 = 0.7 * rcut
        x_g = np.clip((self.r_g - r0) / (rcut - r0), 0, 1)
        f_g = x_g**2 * (3 - 2 * x_g)
        shift = (4 * a_g[gcut] - a_g[gcut - 1]) / 3
        a_g -= f_g * shift
        a_g[gcut + 1:] = 0

    def pseudize_normalized(self, a_g, gc, l=0, points=3):
        """Construct normalized smooth continuation of a_g for g<gc.

        Same as pseudize() with also this constraint::

            /  _     2  /  _     2
            | dr b(r) = | dr a(r)
            /           /
        """

        b_g = self.pseudize(a_g, gc, l, points)[0]
        c_x = np.empty(points + 1)
        gc0 = gc // 2
        x0 = b_g[gc0]
        r_g = self.r_g
        i = [gc0] + list(range(gc, gc + points))
        r_i = r_g[i]
        norm = self.integrate(a_g**2)

        def f(x):
            b_g[gc0] = x[0]
            c_x[:] = np.polyfit(r_i**2, b_g[i] / r_i**l, points)
            b_g[:gc] = np.polyval(c_x, r_g[:gc]**2) * r_g[:gc]**l
            return self.integrate(b_g**2) - norm

        from scipy.optimize import fsolve
        fsolve(f, x0)
        return b_g, c_x[-1]

    def pseudize_smooth(self,
                        a_g: Array1D,
                        gc: int,
                        l: int = 0,
                        points: int = 3,
                        Gcut: float = 6.0) -> Tuple[Array1D, float]:
        """Minimize Fourier components above Gcut."""
        b_g, _ = self.pseudize(a_g, gc, l, points)
        c_x = np.empty(points + 1)
        gc0 = gc // 2
        x0 = b_g[gc0]
        i = [gc0] + list(range(gc, gc + points))
        r_g = self.r_g
        r_i = r_g[i]

        def f(x):
            b_g[gc0] = x[0]
            c_x[:] = np.polyfit(r_i**2, b_g[i] / r_i**l, points)
            b_g[:gc] = np.polyval(c_x, r_g[:gc]**2) * r_g[:gc]**l
            g_k, b_k = self.fft(b_g * r_g, l)
            kc = int(Gcut / g_k[1])
            f_k = g_k[kc:] * b_k[kc:]
            return f_k @ f_k

        from scipy.optimize import minimize
        minimize(f, x0)

        return b_g, c_x[-1]

    def jpseudize(self, a_g, gc, l=0, points=4):
        """Construct spherical Bessel function continuation of a_g for g<gc.

        Returns (b_g, b(0)/r^l) such that b_g=a_g for g >= gc and::

                P-2
            b = Sum c_p j (q r )
             g  p=0      l  p g

        for g < gc+P, where.
        """

        from scipy.optimize import brentq
        from scipy.special import sph_jn

        if a_g[gc] == 0:
            return self.zeros(), 0.0

        assert isinstance(gc, int) and gc > 10

        zeros_l = [[1, 2, 3, 4, 5, 6],
                   [1.430, 2.459, 3.471, 4.477, 5.482, 6.484],
                   [1.835, 2.895, 3.923, 4.938, 5.949, 6.956],
                   [2.224, 3.316, 4.360, 5.387, 6.405, 7.418]]

        # Logarithmic derivative:
        ld = np.dot([-1 / 60, 3 / 20, -3 / 4, 0, 3 / 4, -3 / 20, 1 / 60],
                    a_g[gc - 3:gc + 4]) / a_g[gc] / self.dr_g[gc]

        rc = self.r_g[gc]

        def f(q):
            j, dj = (y[-1] for y in sph_jn(l, q * rc))
            return dj * q - j * ld

        j_pg = self.empty(points - 1)
        q_p = np.empty(points - 1)

        zeros = zeros_l[l]
        if rc * ld > l:
            z1 = zeros[0]
            zeros = zeros[1:]
        else:
            z1 = 0
        x = 0.01
        for p, z2 in enumerate(zeros[:points - 1]):
            q = brentq(f, z1 * pi / rc + x, z2 * pi / rc - x)
            j_pg[p] = [sph_jn(l, q * r)[0][-1] for r in self.r_g]
            q_p[p] = q
            z1 = z2

        C_dg = [[0, 0, 0, 1, 0, 0, 0],
                [1 / 90, -3 / 20, 3 / 2, -49 / 18, 3 / 2, -3 / 20, 1 / 90],
                [1 / 8, -1, 13 / 8, 0, -13 / 8, 1, -1 / 8],
                [-1 / 6, 2, -13 / 2, 28 / 3, -13 / 2, 2, -1 / 6]][:points - 1]
        c_p = np.linalg.solve(np.dot(C_dg, j_pg[:, gc - 3:gc + 4].T),
                              np.dot(C_dg, a_g[gc - 3:gc + 4]))
        b_g = a_g.copy()
        b_g[:gc + 2] = np.dot(c_p, j_pg[:, :gc + 2])
        return b_g, np.dot(c_p, q_p**l) * 2**l * fac(l) / fac(2 * l + 1)

    def plot(self, a_g, n=0, rc=4.0, show=False):
        import matplotlib.pyplot as plt
        r_g = self.r_g[:len(a_g)]
        if n < 0:
            r_g = r_g[1:]
            a_g = a_g[1:]
        plt.plot(r_g, a_g * r_g**n)
        plt.axis(xmin=0, xmax=rc)
        if show:
            plt.show()

    def floor(self, r):
        return np.floor(self.r2g(r)).astype(int)

    def round(self, r):
        return np.around(self.r2g(r)).astype(int)

    def ceil(self, r):
        return np.ceil(self.r2g(r)).astype(int)

    def spline(self, a_g, rcut=None, l=0, points=None):
        """Create spline representation of a radial function f(r).

        The spline represents a rescaled version of the function, f(r) / r^l.
        """
        if points is None:
            points = self.default_spline_points

        if rcut is None:
            # Calculate rcut for the input function, i.e. a value rcut which
            # satisfies f(r) = 0 ∀ r ≥ rcut
            g = len(a_g) - 1
            while a_g[g] == 0.0:
                g -= 1
            rcut = self.r_g[g + 1]

        # Rescale f(r) -> f(r) / r^l
        N = len(a_g)
        b_g = divrl(a_g, l, self.r_g[:N])

        # Interpolate to a uniform radial grid (for the spline representation)
        r_i = np.linspace(0, rcut, points + 1)
        g_i = np.clip((self.r2g(r_i) + 0.5).astype(int), 1, N - 2)
        r1_i = self.r_g[g_i - 1]
        r2_i = self.r_g[g_i]
        r3_i = self.r_g[g_i + 1]
        x1_i = (r_i - r2_i) * (r_i - r3_i) / (r1_i - r2_i) / (r1_i - r3_i)
        x2_i = (r_i - r1_i) * (r_i - r3_i) / (r2_i - r1_i) / (r2_i - r3_i)
        x3_i = (r_i - r1_i) * (r_i - r2_i) / (r3_i - r1_i) / (r3_i - r2_i)
        b1_i = b_g[g_i - 1]
        b2_i = b_g[g_i]
        b3_i = b_g[g_i + 1]
        b_i = b1_i * x1_i + b2_i * x2_i + b3_i * x3_i

        return Spline(l, rcut, b_i)

    def get_cutoff(self, f_g):
        g = self.N - 1
        while f_g[g] == 0.0:
            g -= 1
        gcut = g + 1
        return gcut

    @abstractmethod
    def r2g(self, r):
        """Inverse continuous map from a real space distance (r)
           to a floating point grid index (g).

           Used by methods floor, round, and ceil, which manipulate this
           floating point to an integer accordingly.
        """


class EquidistantRadialGridDescriptor(RadialGridDescriptor):
    def __init__(self, h, N=1000, h0=0.0):
        """Equidistant radial grid descriptor.

        The radial grid is r(g) = h0 + g*h,  g = 0, 1, ..., N - 1."""

        RadialGridDescriptor.__init__(self,
                                      h * np.arange(N) + h0,
                                      h + np.zeros(N))

    def r2g(self, r):
        return int(np.ceil((r - self.r_g[0]) / (self.r_g[1] - self.r_g[0])))

    def xml(self, id='grid1'):
        assert self.r_g[0] == 0.0
        return ('<radial_grid eq="r=d*i" d="{0!r}" '
                'istart="0" iend="{1}" id="{2}"/>\n'
                .format(self.r_g[1], len(self.r_g) - 1, id))

    def spline(self, a_g, rcut=None, l=0, points=None):
        b_g = a_g.copy()
        if l > 0:
            b_g = divrl(b_g, l, self.r_g[:len(a_g)])
        return Spline(l, self.r_g[len(a_g) - 1], b_g)


class AERadialGridDescriptor(RadialGridDescriptor):
    def __init__(self, a, b, N=1000, default_spline_points=25):
        """Radial grid descriptor for all-electron calculation.

        The radial grid is::

                     a g
            r(g) = -------,  g = 0, 1, ..., N - 1
                   1 - b g
        """

        self.a = a
        self.b = b
        g = np.arange(N)
        r_g = self.a * g / (1 - self.b * g)
        dr_g = (self.b * r_g + self.a)**2 / self.a
        RadialGridDescriptor.__init__(self, r_g, dr_g, default_spline_points)

    def r2g(self, r):
        # return r / (r * self.b + self.a)
        # Hack to preserve backwards compatibility:
        ng = 1.0 / self.b
        beta = self.a / self.b
        return ng * r / (beta + r)

    def new(self, N):
        return AERadialGridDescriptor(self.a, self.b, N)

    def xml(self, id='grid1'):
        if abs(self.N - 1 / self.b) < 1e-5:
            return (('<radial_grid eq="r=a*i/(n-i)" a="%r" n="%d" ' +
                     'istart="0" iend="%d" id="%s"/>\n') %
                    (self.a * self.N, self.N, self.N - 1, id))
        return (('<radial_grid eq="r=a*i/(1-b*i)" a="%r" b="%r" n="%d" ' +
                 'istart="0" iend="%d" id="%s"/>\n') %
                (self.a, self.b, self.N, self.N - 1, id))

    def d2gdr2(self):
        return -2 * self.a * self.b / (self.b * self.r_g + self.a)**3


class AbinitRadialGridDescriptor(RadialGridDescriptor):
    def __init__(self, a, d, N=1000, default_spline_points=25):
        """Radial grid descriptor for Abinit calculations.

        The radial grid is::

                      dg
            r(g) = a(e   - 1),  g = 0, 1, ..., N - 1
        """

        self.a = a
        self.d = d
        g = np.arange(N)
        r_g = a * (np.exp(d * g) - 1)
        dr_g = (r_g + a) * d
        RadialGridDescriptor.__init__(self, r_g, dr_g, default_spline_points)

    def r2g(self, r):
        return np.log(r / self.a + 1) / self.d

    def d2gdr2(self):
        return -1 / (self.a**2 * self.d * (self.r_g / self.a + 1)**2)

    def new(self, N):
        return AbinitRadialGridDescriptor(self.a, self.d, N)



import numpy as np
import pytest
from ase.build import molecule
from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.mpi import world
from gpaw.poisson import PoissonSolver
from gpaw.tddft.spectrum import photoabsorption_spectrum
from gpaw.test import equal


@pytest.mark.rttddft
def test_lcaotddft_simple(in_tmp_dir):
    atoms = molecule('Na2')
    atoms.center(vacuum=4.0)

    # Ground-state calculation
    calc = GPAW(nbands=2, h=0.4, setups=dict(Na='1'),
                basis='dzp', mode='lcao',
                poissonsolver=PoissonSolver('fd', eps=1e-16),
                convergence={'density': 1e-8},
                txt='gs.out',
                symmetry={'point_group': False})
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Time-propagation calculation
    td_calc = LCAOTDDFT('gs.gpw', txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick(np.ones(3) * 1e-5)
    td_calc.propagate(20, 3)
    photoabsorption_spectrum('dm.dat', 'spec.dat', delta_e=5)
    world.barrier()

    # Test dipole moment
    data_i = np.loadtxt('dm.dat')[:, 2:].ravel()
    if 0:
        from gpaw.test import print_reference
        print_reference(data_i, 'ref_i', '%.12le')

    ref_i = [-9.383700894739e-16,
             -9.338586948130e-16,
             2.131582675483e-14,
             8.679923327633e-15,
             7.529517689096e-15,
             2.074867751820e-14,
             1.967175558125e-05,
             1.967175557952e-05,
             1.805004256446e-05,
             3.799528978877e-05,
             3.799528978943e-05,
             3.602506734201e-05,
             5.371491974467e-05,
             5.371491974534e-05,
             5.385046706407e-05]

    tol = 1e-10
    equal(data_i, ref_i, tol)

    # Test spectrum
    data_i = np.loadtxt('spec.dat').ravel()
    if 0:
        from gpaw.test import print_reference
        print_reference(data_i, 'ref_i', '%.12le')

    ref_i = [0.000000000000e+00,
             0.000000000000e+00,
             0.000000000000e+00,
             0.000000000000e+00,
             5.000000000000e+00,
             4.500226856200e-03,
             4.500226856200e-03,
             4.408379542600e-03,
             1.000000000000e+01,
             1.659426124300e-02,
             1.659426124300e-02,
             1.623812256800e-02,
             1.500000000000e+01,
             3.244686838800e-02,
             3.244686838800e-02,
             3.168682490900e-02,
             2.000000000000e+01,
             4.684883744600e-02,
             4.684883744600e-02,
             4.559689861200e-02,
             2.500000000000e+01,
             5.466781222200e-02,
             5.466781222200e-02,
             5.290171209500e-02,
             3.000000000000e+01,
             5.231586230700e-02,
             5.231586230700e-02,
             5.008661764300e-02]

    tol = 1e-7
    equal(data_i, ref_i, tol)


@pytest.mark.rttddft
def test_lcaotddft_fail_with_symmetry(in_tmp_dir):
    atoms = molecule('Na2')
    atoms.center(vacuum=4.0)

    # Ground-state calculation with point group symmetries
    calc = GPAW(nbands=2, h=0.4, setups=dict(Na='1'),
                basis='dzp', mode='lcao',
                poissonsolver=PoissonSolver('fd', eps=1e-16),
                convergence={'density': 1e-8},
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Time-propagation calculation
    # should not be allowed with symmetries
    with pytest.raises(ValueError):
        LCAOTDDFT('gs.gpw', txt='td.out')



import pytest
from ase import Atoms

from gpaw import GPAW, LCAO
from gpaw.poisson import FDPoissonSolver
from gpaw.test import equal


@pytest.mark.later  # basis set cutoff??
def test_lcao_h2o():
    a = 6.0
    b = a / 2
    mol = Atoms('OHH',
                [(b, b, 0.1219 + b),
                 (b, 0.7633 + b, -0.4876 + b),
                 (b, -0.7633 + b, -0.4876 + b)],
                pbc=False, cell=[a, a, a])
    calc = GPAW(gpts=(32, 32, 32),
                nbands=4,
                mode='lcao',
                poissonsolver=FDPoissonSolver())
    mol.calc = calc
    e = mol.get_potential_energy()
    niter = calc.get_number_of_iterations()

    equal(e, -10.271, 2e-3)
    equal(niter, 8, 1)

    # Check that complex wave functions are allowed with
    # gamma point calculations
    calc = GPAW(gpts=(32, 32, 32),
                nbands=4,
                mode=LCAO(force_complex_dtype=True),
                poissonsolver=FDPoissonSolver())
    mol.calc = calc
    ec = mol.get_potential_energy()
    equal(e, ec, 1e-5)



from math import sqrt

import numpy as np
import pytest
from ase import Atoms

from gpaw import GPAW, PW, restart
from gpaw.test import equal


@pytest.mark.later
def test_fileio_restart(in_tmp_dir):
    d = 3.0
    atoms = Atoms('Na3',
                  positions=[(0, 0, 0),
                             (0, 0, d),
                             (0, d * sqrt(3 / 4), d / 2)],
                  magmoms=[1.0, 1.0, 1.0],
                  cell=(3.5, 3.5, 4 + 2 / 3),
                  pbc=True)

    def test(atoms):
        e0 = atoms.get_potential_energy()
        f0 = atoms.get_forces()
        m0 = atoms.get_magnetic_moments()
        eig00 = atoms.calc.get_eigenvalues(spin=0)
        eig01 = atoms.calc.get_eigenvalues(spin=1)
        wf0 = atoms.calc.get_pseudo_wave_function(band=1)
        # Write the restart file(s):
        atoms.calc.write('tmp1.gpw')
        atoms.calc.write('tmp2.gpw', 'all')

        # Try restarting:
        atoms, calc = restart('tmp2.gpw', txt=None)
        wf1 = calc.get_pseudo_wave_function(band=1)
        e1 = atoms.get_potential_energy()
        f1 = atoms.get_forces()
        m1 = atoms.get_magnetic_moments()
        eig10 = calc.get_eigenvalues(spin=0)
        eig11 = calc.get_eigenvalues(spin=1)
        print(e0, e1)
        equal(e0, e1, 1e-10)
        print(f0, f1)
        for ff0, ff1 in zip(f0, f1):
            err = np.linalg.norm(ff0 - ff1)
            assert err <= 1e-10
        print(m0, m1)
        for mm0, mm1 in zip(m0, m1):
            equal(mm0, mm1, 1e-10)
        print('A', eig00, eig10)
        for eig0, eig1 in zip(eig00, eig10):
            equal(eig0, eig1, 1e-10)
        print('B', eig01, eig11)
        for eig0, eig1 in zip(eig01, eig11):
            equal(eig0, eig1, 1e-10)
        equal(abs(wf1 - wf0).max(), 0, 1e-14)

        # Check that after restart everything is writable
        calc.write('tmp3.gpw')
        calc.write('tmp4.gpw', 'all')

    # Only a short, non-converged calculation
    conv = {'eigenstates': 1.24, 'energy': 2e-1, 'density': 1e-1}

    for kwargs in [{'mode': PW(200)},
                   {'mode': 'fd', 'h': 0.30}]:
        atoms.calc = GPAW(nbands=3,
                          setups={'Na': '1'},
                          convergence=conv,
                          txt=None,
                          **kwargs)
        test(atoms)



import pytest
from gpaw.broadcast_imports import broadcast_imports

# These tests would be better if we could find a way to call them
# without first executing gpaw.__init__ (e.g. with pytest-forked or
# something?).
#
# However at least the bad() test is useful because it specifies the
# fail behaviour which may not otherwise be exercised.
#
# Some day: Execute the imports in such a way that we can verify the
# module counts.


def test_bcast_imports_good():
    with broadcast_imports:
        import gpaw.poisson as poisson_module
        assert poisson_module is not None


def test_bcast_imports_bad():
    with broadcast_imports:
        with pytest.raises(ModuleNotFoundError):
            import gpaw.module_not_to_be_found  # noqa



from math import sqrt, pi

import numpy as np
import pytest

from gpaw.setup import create_setup
from gpaw.grid_descriptor import GridDescriptor
from gpaw.lfc import LFC
from gpaw.xc import XC

n = 60  # 40 /8 * 10
a = 10.0


@pytest.mark.ci
def test_multipole():
    gd = GridDescriptor((n, n, n), (a, a, a))
    c_LL = np.identity(9, float)
    a_Lg = gd.zeros(9)
    xc = XC('LDA')
    for soft in [False]:
        s = create_setup('Cu', xc, lmax=2)
        ghat_l = s.ghat_l
        ghat_Lg = LFC(gd, [ghat_l])
        ghat_Lg.set_positions([(0.54321, 0.5432, 0.543)])
        a_Lg[:] = 0.0
        ghat_Lg.add(a_Lg, {0: c_LL} if ghat_Lg.my_atom_indices else {})
        for l in range(3):
            for m in range(2 * l + 1):
                L = l**2 + m
                a_g = a_Lg[L]
                Q0 = gd.integrate(a_g) / sqrt(4 * pi)
                Q1_m = -gd.calculate_dipole_moment(a_g) / sqrt(4 * pi / 3)
                print(Q0)
                if l == 0:
                    Q0 -= 1.0
                    Q1_m[:] = 0.0
                elif l == 1:
                    Q1_m[(m + 1) % 3] -= 1.0
                print(soft, l, m, Q0, Q1_m)
                assert abs(Q0) < 2e-6
                assert (abs(Q1_m) < 3e-5).all()
        b_Lg = np.reshape(a_Lg, (9, -1))
        S_LL = np.inner(b_Lg, b_Lg)
        gd.comm.sum(S_LL)
        S_LL.ravel()[::10] = 0.0
        print(max(abs(S_LL).ravel()))
        assert max(abs(S_LL).ravel()) < 3e-4



import re
import numpy as np

from gpaw import __version__ as version
from gpaw.mpi import world
from gpaw.tddft.units import au_to_as, au_to_fs, au_to_eV, rot_au_to_cgs
from gpaw.tddft.folding import FoldedFrequencies
from gpaw.tddft.folding import Folding


def calculate_fourier_transform(x_t, y_ti, foldedfrequencies):
    ff = foldedfrequencies
    X_w = ff.frequencies
    envelope = ff.folding.envelope

    # Construct integration weights:
    # We use trapezoidal rule except the end point is accounted with
    # full weight. This ensures better numerical compatibility
    # with the on-the-fly Fourier integrators.
    # This shouldn't affect in usual scenarios as the envelope
    # should damp the data to zero at the end point in any case.
    dx_t1 = x_t[1:] - x_t[:-1]
    dx_t = 0.5 * (np.insert(dx_t1, 0, 0.0) + np.append(dx_t1, dx_t1[-1]))

    # Integrate
    f_wt = np.exp(1.0j * np.outer(X_w, x_t))
    y_it = np.swapaxes(y_ti, 0, 1)
    Y_wi = np.tensordot(f_wt, dx_t * envelope(x_t) * y_it, axes=(1, 1))
    return Y_wi


def read_td_file_data(fname, remove_duplicates=True):
    """Read data from time-dependent data file.

    Parameters
    ----------
    fname
        File path
    remove_duplicates
        If true, remove data from overlapping time values.
        The first encountered values are kept.

    Returns
    -------
    time_t
        Array of time values
    data_ti
        Array of data values
    """
    # Read data
    data_tj = np.loadtxt(fname)
    time_t = data_tj[:, 0]
    data_ti = data_tj[:, 1:]

    # Remove duplicates due to abruptly stopped and restarted calculation
    if remove_duplicates:
        flt_t = np.ones_like(time_t, dtype=bool)
        maxtime = time_t[0]
        for t in range(1, len(time_t)):
            # Note about ">=" here:
            # The equality is included here in order to
            # retain step-like data (for example, the data both just before
            # and just after the kick is kept).
            if time_t[t] >= maxtime:
                maxtime = time_t[t]
            else:
                flt_t[t] = False
        time_t = time_t[flt_t]
        data_ti = data_ti[flt_t]
        ndup = len(flt_t) - flt_t.sum()
        if ndup > 0:
            print('Removed %d duplicates' % ndup)
    return time_t, data_ti


def read_td_file_kicks(fname):
    """Read kicks from time-dependent data file.

    Parameters
    ----------
    fname
        File path

    Returns
    -------
    kick_i
        List of kicks.
        Each kick is a dictionary with keys
        ``strength_v`` and ``time``.
    """
    def parse_kick_line(line):
        # Kick
        regexp = (r"Kick = \["
                  r"(?P<k0>[-+0-9\.e\ ]+), "
                  r"(?P<k1>[-+0-9\.e\ ]+), "
                  r"(?P<k2>[-+0-9\.e\ ]+)\]")
        m = re.search(regexp, line)
        assert m is not None, 'Kick not found'
        kick_v = np.array([float(m.group('k%d' % v)) for v in range(3)])
        # Time
        regexp = r"Time = (?P<time>[-+0-9\.e\ ]+)"
        m = re.search(regexp, line)
        if m is None:
            print('time not found')
            time = 0.0
        else:
            time = float(m.group('time'))
        return kick_v, time

    # Search kicks
    kick_i = []
    with open(fname, 'r') as f:
        for line in f:
            if line.startswith('# Kick'):
                kick_v, time = parse_kick_line(line)
                kick_i.append({'strength_v': kick_v, 'time': time})
    return kick_i


def clean_td_data(kick_i, time_t, data_ti):
    """Prune time-dependent data.

    This function checks that there is only one kick
    in the kick list and moves time zero to the kick
    time (discarding all preceding data).

    Parameters
    ----------
    kick_i
        List of kicks.
    time_t
        Array of time values
    data_ti
        Array of data values

    Returns
    -------
    kick_i
        List of kicks.
        Each kick is a dictionary with keys
        ``strength_v`` and ``time``.

    Raises
    ------
    RuntimeError
        If kick list contains multiple kicks.
    """
    # Check kicks
    if len(kick_i) > 1:
        raise RuntimeError('Multiple kicks')
    kick = kick_i[0]
    kick_v = kick['strength_v']
    kick_time = kick['time']

    # Discard times before kick
    flt_t = time_t >= kick_time
    time_t = time_t[flt_t]
    data_ti = data_ti[flt_t]

    # Move time zero to kick time
    time_t -= kick_time
    assert time_t[0] == 0.0

    return kick_v, time_t, data_ti


def read_dipole_moment_file(fname, remove_duplicates=True):
    """Read time-dependent dipole moment data file.

    Parameters
    ----------
    fname
        File path
    remove_duplicates
        If true, remove data from overlapping time values.
        The first encountered values are kept.

    Returns
    -------
    kick_i
        List of kicks.
        Each kick is a dictionary with keys
        ``strength_v`` and ``time``.
    time_t
        Array of time values
    norm_t
        Array of norm values
    dm_tv
        Array of dipole moment values
    """
    time_t, data_ti = read_td_file_data(fname, remove_duplicates)
    kick_i = read_td_file_kicks(fname)
    norm_t = data_ti[:, 0]
    dm_tv = data_ti[:, 1:]
    return kick_i, time_t, norm_t, dm_tv


def calculate_polarizability(kick_v, time_t, dm_tv, foldedfrequencies):
    dm_tv = dm_tv - dm_tv[0]
    alpha_wv = calculate_fourier_transform(time_t, dm_tv, foldedfrequencies)
    kick_magnitude = np.sqrt(np.sum(kick_v**2))
    alpha_wv /= kick_magnitude
    return alpha_wv


def calculate_photoabsorption(kick_v, time_t, dm_tv, foldedfrequencies):
    omega_w = foldedfrequencies.frequencies
    alpha_wv = calculate_polarizability(kick_v, time_t, dm_tv,
                                        foldedfrequencies)
    abs_wv = 2 / np.pi * omega_w[:, np.newaxis] * alpha_wv.imag

    kick_magnitude = np.sqrt(np.sum(kick_v**2))
    abs_wv *= kick_v / kick_magnitude
    return abs_wv


def read_magnetic_moment_file(fname, remove_duplicates=True):
    """Read time-dependent magnetic moment data file.

    Parameters
    ----------
    fname
        File path
    remove_duplicates
        If true, remove data from overlapping time values.
        The first encountered values are kept.

    Returns
    -------
    kick_i
        List of kicks.
        Each kick is a dictionary with keys
        ``strength_v`` and ``time``.
    time_t
        Array of time values
    mm_tv
        Array of magnetic moment values
    """
    time_t, mm_tv = read_td_file_data(fname, remove_duplicates)
    kick_i = read_td_file_kicks(fname)
    return kick_i, time_t, mm_tv


def calculate_rotatory_strength_components(kick_v, time_t, mm_tv,
                                           foldedfrequencies):
    assert np.all(mm_tv[0] == 0.0)
    mm_wv = calculate_fourier_transform(time_t, mm_tv, foldedfrequencies)
    kick_magnitude = np.sqrt(np.sum(kick_v**2))
    rot_wv = mm_wv.real / (np.pi * kick_magnitude)
    return rot_wv


def write_spectrum(dipole_moment_file, spectrum_file,
                   folding, width, e_min, e_max, delta_e,
                   title, symbol, calculate):
    def str_list(v_i, fmt='%g'):
        return '[%s]' % ', '.join(map(lambda v: fmt % v, v_i))

    kick_i, time_t, _, dm_tv = read_dipole_moment_file(dipole_moment_file)
    kick_v, time_t, dm_tv = clean_td_data(kick_i, time_t, dm_tv)
    dt_t = time_t[1:] - time_t[:-1]

    freqs = np.arange(e_min, e_max + 0.5 * delta_e, delta_e)
    folding = Folding(folding, width)
    ff = FoldedFrequencies(freqs, folding)
    omega_w = ff.frequencies
    spec_wv = calculate(kick_v, time_t, dm_tv, ff)

    # Write spectrum file header
    with open(spectrum_file, 'w') as f:
        def w(s):
            f.write('%s\n' % s)

        w('# %s spectrum from real-time propagation' % title)
        w('# GPAW version: %s' % version)
        w('# Total time = %.4f fs, Time steps = %s as' %
          (dt_t.sum() * au_to_fs,
           str_list(np.unique(np.around(dt_t, 6)) * au_to_as, '%.4f')))
        w('# Kick = %s' % str_list(kick_v))
        w('# %sian folding, Width = %.4f eV = %lf Hartree'
          ' <=> FWHM = %lf eV' %
          (folding.folding, folding.width * au_to_eV, folding.width,
           folding.fwhm * au_to_eV))

        col_i = []
        data_iw = [omega_w * au_to_eV]
        for v in range(len(kick_v)):
            h = '%s_%s' % (symbol, 'xyz'[v])
            if spec_wv.dtype == complex:
                col_i.append('Re[%s]' % h)
                data_iw.append(spec_wv[:, v].real)
                col_i.append('Im[%s]' % h)
                data_iw.append(spec_wv[:, v].imag)
            else:
                col_i.append(h)
                data_iw.append(spec_wv[:, v])

        w('# %10s %s' % ('om (eV)', ' '.join(['%20s' % s for s in col_i])))

    # Write spectrum file data
    with open(spectrum_file, 'ab') as f:
        np.savetxt(f, np.array(data_iw).T,
                   fmt='%12.6lf' + (' %20.10le' * len(col_i)))

    return folding.envelope(time_t[-1])


def photoabsorption_spectrum(dipole_moment_file: str,
                             spectrum_file: str,
                             folding: str = 'Gauss',
                             width: float = 0.2123,
                             e_min: float = 0.0,
                             e_max: float = 30.0,
                             delta_e: float = 0.05):
    """Calculates photoabsorption spectrum from the time-dependent
    dipole moment.

    The spectrum is represented as a dipole strength function
    in units of 1/eV. Thus, the resulting spectrum should integrate
    to the number of valence electrons in the system.

    Parameters
    ----------
    dipole_moment_file
        Name of the time-dependent dipole moment file from which
        the spectrum is calculated
    spectrum_file
        Name of the spectrum file
    folding
        Gaussian (``'Gauss'``) or Lorentzian (``'Lorentz'``) folding
    width
        Width of the Gaussian (sigma) or Lorentzian (Gamma)
        Gaussian =     1/(sigma sqrt(2pi)) exp(-(1/2)(omega/sigma)^2)
        Lorentzian =  (1/pi) (1/2) Gamma / [omega^2 + ((1/2) Gamma)^2]
    e_min
        Minimum energy shown in the spectrum (eV)
    e_max
        Maximum energy shown in the spectrum (eV)
    delta_e
        Energy resolution (eV)
    """
    if world.rank == 0:
        print('Calculating photoabsorption spectrum from file "%s"'
              % dipole_moment_file)

        def calculate(*args):
            return calculate_photoabsorption(*args) / au_to_eV
        sinc = write_spectrum(dipole_moment_file, spectrum_file,
                              folding, width, e_min, e_max, delta_e,
                              'Photoabsorption', 'S', calculate)
        print('Sinc contamination %.8f' % sinc)
        print('Calculated photoabsorption spectrum saved to file "%s"'
              % spectrum_file)


def polarizability_spectrum(dipole_moment_file, spectrum_file,
                            folding='Gauss', width=0.2123,
                            e_min=0.0, e_max=30.0, delta_e=0.05):
    """Calculates polarizability spectrum from the time-dependent
    dipole moment.

    Parameters:

    dipole_moment_file: string
        Name of the time-dependent dipole moment file from which
        the spectrum is calculated
    spectrum_file: string
        Name of the spectrum file
    folding: 'Gauss' or 'Lorentz'
        Whether to use Gaussian or Lorentzian folding
    width: float
        Width of the Gaussian (sigma) or Lorentzian (Gamma)
        Gaussian =     1/(sigma sqrt(2pi)) exp(-(1/2)(omega/sigma)^2)
        Lorentzian =  (1/pi) (1/2) Gamma / [omega^2 + ((1/2) Gamma)^2]
    e_min: float
        Minimum energy shown in the spectrum (eV)
    e_max: float
        Maximum energy shown in the spectrum (eV)
    delta_e: float
        Energy resolution (eV)
    """
    if world.rank == 0:
        print('Calculating polarizability spectrum from file "%s"'
              % dipole_moment_file)

        def calculate(*args):
            return calculate_polarizability(*args) / au_to_eV**2
        sinc = write_spectrum(dipole_moment_file, spectrum_file,
                              folding, width, e_min, e_max, delta_e,
                              'Polarizability', 'alpha', calculate)
        print('Sinc contamination %.8f' % sinc)
        print('Calculated polarizability spectrum saved to file "%s"'
              % spectrum_file)


def rotatory_strength_spectrum(magnetic_moment_files, spectrum_file,
                               folding='Gauss', width=0.2123,
                               e_min=0.0, e_max=30.0, delta_e=0.05):
    """Calculates rotatory strength spectrum from the time-dependent
    magnetic moment.

    Parameters
    ----------
    magnetic_moment_files: list of string
        Time-dependent magnetic moment files for x, y, and z kicks
    spectrum_file: string
        Name of the spectrum file
    folding: 'Gauss' or 'Lorentz'
        Whether to use Gaussian or Lorentzian folding
    width: float
        Width of the Gaussian (sigma) or Lorentzian (Gamma)
        Gaussian =     1/(sigma sqrt(2pi)) exp(-(1/2)(omega/sigma)^2)
        Lorentzian =  (1/pi) (1/2) Gamma / [omega^2 + ((1/2) Gamma)^2]
    e_min: float
        Minimum energy shown in the spectrum (eV)
    e_max: float
        Maximum energy shown in the spectrum (eV)
    delta_e: float
        Energy resolution (eV)
    """
    if world.rank != 0:
        return

    freqs = np.arange(e_min, e_max + 0.5 * delta_e, delta_e)
    folding = Folding(folding, width)
    ff = FoldedFrequencies(freqs, folding)
    omega_w = ff.frequencies * au_to_eV
    rot_w = np.zeros_like(omega_w)

    tot_time = np.inf
    time_steps = []
    kick_strength = None
    for v, fpath in enumerate(magnetic_moment_files):
        kick_i, time_t, mm_tv = read_magnetic_moment_file(fpath)
        kick_v, time_t, mm_tv = clean_td_data(kick_i, time_t, mm_tv)

        tot_time = min(tot_time, time_t[-1])
        time_steps.append(np.around(time_t[1:] - time_t[:-1], 6))

        # Check kicks
        for v0 in range(3):
            if v0 == v:
                continue
            if kick_v[v0] != 0.0:
                raise RuntimeError('The magnetic moment files must be '
                                   'for kicks in x, y, and z directions.')
        if kick_strength is None:
            kick_strength = np.sqrt(np.sum(kick_v**2))
        if np.sqrt(np.sum(kick_v**2)) != kick_strength:
            raise RuntimeError('The magnetic moment files must have '
                               'been calculated with the same kick strength.')

        rot_wv = calculate_rotatory_strength_components(kick_v, time_t,
                                                        mm_tv, ff)
        rot_w += rot_wv[:, v]

    rot_w *= rot_au_to_cgs * 1e40 / au_to_eV

    # Unique non-zero time steps
    time_steps = np.unique(time_steps)
    time_steps = time_steps[time_steps != 0]

    with open(spectrum_file, 'w') as fd:
        steps_str = ', '.join(f'{val:.4f}' for val in time_steps * au_to_as)
        lines = ['Rotatory strength spectrum from real-time propagations',
                 f'Total time = {tot_time * au_to_fs:.4f} fs, '
                 f'Time steps = [{steps_str}] as',
                 f'Kick strength = {kick_strength}',
                 f'{folding.folding}ian folding, '
                 f'width = {folding.width * au_to_eV:.4f} eV '
                 f'<=> FWHM = {folding.fwhm * au_to_eV:.4f} eV']
        fd.write('# ' + '\n# '.join(lines) + '\n')
        fd.write(f'# {"Energy (eV)":>12} {"R (1e-40 cgs / eV)":>20}\n')

        data_wi = np.vstack((omega_w, rot_w)).T
        np.savetxt(fd, data_wi,
                   fmt='%14.6lf' + (' %20.10le' * (data_wi.shape[1] - 1)))



import numpy as np
import pytest

from ase import Atoms
from ase.units import Ha

from gpaw import GPAW
from gpaw.convergence_criteria import WorkFunction, Energy, Criterion


class FourIterations(Criterion):
    """A silly custom convergence criterion that ensures it runs
    at least four iterations."""
    name = 'four iterations'
    tablename = 'four'
    calc_last = False

    def __init__(self):
        self.description = 'At least four iterations must complete!'
        self.reset()

    def __call__(self, context):
        converged = self.iters >= 4
        entry = str(self.iters)
        self.iters += 1
        return converged, entry

    def reset(self):
        self.iters = 0


@pytest.mark.later
def test_scf_criterion(in_tmp_dir):
    """Tests different ways of setting SCF convergence criteria,
    and that it behaves consistenly with regard to the work function."""
    convergence = {'eigenstates': 1.0,
                   'density': 1.0,
                   'energy': 1.0,
                   'work function': 1.0}
    atoms = Atoms('HF', [(0., 0.5, 0.5),
                         (0., 0.4, -0.4)],
                  cell=(5., 5., 9.),
                  pbc=(True, True, False))
    atoms.center()
    atoms.calc = GPAW(mode='fd',
                      h=0.3,
                      nbands=-1,
                      convergence=convergence,
                      txt=None,
                      poissonsolver={'dipolelayer': 'xy'})
    atoms.get_potential_energy()
    workfunctions1 = Ha * atoms.calc.hamiltonian.get_workfunctions(
        atoms.calc.wfs)
    atoms.calc.write('scf-criterion.gpw')

    # Flip and use saved calculator; work functions should be opposite.
    atoms = Atoms('HF', [(0., 0.5, -0.5),
                         (0., 0.4, +0.4)],
                  cell=(5., 5., 9.),
                  pbc=(True, True, False))
    atoms.center()
    atoms.calc = GPAW('scf-criterion.gpw', txt=None)  # checks loading
    atoms.get_potential_energy()
    workfunctions2 = Ha * atoms.calc.hamiltonian.get_workfunctions(
        atoms.calc.wfs)

    assert workfunctions1[0] == pytest.approx(workfunctions2[1])
    assert workfunctions1[1] == pytest.approx(workfunctions2[0])
    assert atoms.calc.scf.criteria['work function'].tol == pytest.approx(1.0)

    # Try import syntax, and verify it creates a new instance internally.
    workfunction = WorkFunction(0.5)
    convergence = {'eigenstates': 1.0,
                   'density': 1.0,
                   'energy': 1.0,
                   'work function': workfunction}
    atoms.calc = atoms.calc.new(convergence=convergence)
    atoms.get_potential_energy()
    assert atoms.calc.scf.criteria['work function'] is not workfunction
    assert atoms.calc.scf.criteria['work function'].tol == pytest.approx(0.5)

    # Switch to H2 for faster calcs.
    for atom in atoms:
        atom.symbol = 'H'

    # Change a default.
    convergence = {'energy': Energy(2.0, n_old=4),
                   'density': np.inf,
                   'eigenstates': np.inf}
    atoms.calc = atoms.calc.new(convergence=convergence)
    atoms.get_potential_energy()
    assert atoms.calc.scf.criteria['energy'].n_old == 4


@pytest.mark.later
def test_scf_custom_criterion(in_tmp_dir):
    """Simulate a user creating their own custom convergence criterion,
    saving the .gpw file, and re-loading it. It will warn the user at two
    points."""
    convergence = {'eigenstates': 1.0,
                   'density': 1.0,
                   'energy': 1.0,
                   'custom': FourIterations()}

    atoms = Atoms('HF', [(0., 0.5, 0.5),
                         (0., 0.4, -0.4)],
                  cell=(5., 5., 9.),
                  pbc=(True, True, False))
    atoms.center()
    atoms.rattle()
    calc = GPAW(mode='fd',
                h=0.3,
                nbands=-1,
                convergence=convergence,
                txt='out.txt',
                poissonsolver={'dipolelayer': 'xy'})
    atoms.calc = calc
    with pytest.warns(UserWarning):
        # Warns the user that their criterion must be a unique instance,
        # and can't be saved/loaded.
        atoms.get_potential_energy()
    calc.write('four.gpw')
    with pytest.warns(UserWarning):
        # Warns the user that their criterion did not load.
        calc = GPAW('four.gpw', txt='out2.txt')
    atoms[1].x += 0.1
    atoms.calc = calc
    atoms.get_potential_energy()



import numpy as np
import pytest

from ase.units import Bohr
from gpaw.poisson import PoissonSolver, NoInteractionPoissonSolver
from gpaw.poisson_moment import MomentCorrectionPoissonSolver, MomentCorrection
from gpaw.poisson_extravacuum import ExtraVacuumPoissonSolver
from gpaw.grid_descriptor import GridDescriptor

from gpaw.test import equal


@pytest.mark.parametrize('moment_corrections, expected_len', [
    (None, 0),
    ([], 0),
    (4, 1),
    (9, 1),
    ([dict(moms=range(4), center=np.array([1, 3, 5]))], 1),
    ([dict(moms=range(4), center=np.array([5, 3, 5])),
     dict(moms=range(4), center=np.array([7, 5, 3]))], 2)
])
def test_defaults(moment_corrections, expected_len):
    poisson_ref = NoInteractionPoissonSolver()
    poisson = MomentCorrectionPoissonSolver(
        poissonsolver=poisson_ref,
        moment_corrections=moment_corrections)

    assert isinstance(poisson.moment_corrections, list), \
        poisson.moment_corrections
    assert len(poisson.moment_corrections) == expected_len
    assert all([isinstance(mom, MomentCorrection)
                for mom in poisson.moment_corrections])


@pytest.mark.parametrize('moment_corrections', [
    None,
    [],
])
def test_description_empty(moment_corrections):
    poisson_ref = NoInteractionPoissonSolver()
    poisson = MomentCorrectionPoissonSolver(
        poissonsolver=poisson_ref,
        moment_corrections=moment_corrections)

    desc = poisson.get_description()
    desc_ref = poisson_ref.get_description()

    assert isinstance(desc, str)
    assert isinstance(desc_ref, str)
    assert desc_ref in desc
    assert '0 moment corrections' in desc


@pytest.mark.parametrize('moment_corrections, expected_strings', [
    (4, ['1 moment corrections', 'center', 'range(0, 4)']),
    (9, ['1 moment corrections', 'center', 'range(0, 9)']),
    ([dict(moms=range(4), center=np.array([1, 1, 1]))],
        ['1 moment corrections', '[1.00, 1.00, 1.00]', 'range(0, 4)']),
    ([dict(moms=[1, 2, 3], center=np.array([1, 1, 1]))],
        ['1 moment corrections', '[1.00, 1.00, 1.00]', 'range(1, 4)']),
    ([dict(moms=[0, 2, 3], center=np.array([1, 1, 1]))],
        ['1 moment corrections', '[1.00, 1.00, 1.00]', '(0, 2, 3)']),
    ([dict(moms=range(4), center=np.array([2, 3, 4])),
      dict(moms=range(4), center=np.array([7.4, 3.1, 0.1]))],
        ['2 moment corrections', '[2.00, 3.00, 4.00]',
         '[7.40, 3.10, 0.10]', 'range(0, 4)']),
])
def test_description(moment_corrections, expected_strings):
    poisson_ref = NoInteractionPoissonSolver()
    poisson = MomentCorrectionPoissonSolver(
        poissonsolver=poisson_ref,
        moment_corrections=moment_corrections)

    desc = poisson.get_description()
    desc_ref = poisson_ref.get_description()

    assert isinstance(desc, str)
    assert isinstance(desc_ref, str)

    # Make sure that the description starts with the description of the wrapped
    # solver
    assert desc.startswith(desc_ref)

    # and follows with the moments
    desc_rem = desc[len(desc_ref):]
    for expected_str in expected_strings:
        assert expected_str in desc_rem, \
            f'"{expected_str}" not in "{desc_rem}"'


@pytest.mark.parametrize('moment_corrections, expected_string', [
    ([], 'no corrections'),
    (4, 'array([0, 1, 2, 3]) @ None'),
    (9, 'array([0, 1, 2, 3, 4, 5, 6, 7, 8]) @ None'),
    ([dict(moms=range(4), center=np.array([1., 1., 1.]))],
        'array([0, 1, 2, 3]) @ array([1., 1., 1.])'),
    ([dict(moms=[1, 2, 3], center=np.array([1., 1., 1.]))],
        'array([1, 2, 3]) @ array([1., 1., 1.])'),
    ([dict(moms=[0, 2, 3], center=np.array([1., 1., 1.]))],
        'array([0, 2, 3]) @ array([1., 1., 1.])'),
    ([dict(moms=range(4), center=np.array([2, 3, 4])),
      dict(moms=range(4), center=np.array([7.4, 3.1, 0.1]))],
        '2 corrections'),
])
def test_repr(moment_corrections, expected_string):
    poisson_ref = NoInteractionPoissonSolver()
    poisson = MomentCorrectionPoissonSolver(
        poissonsolver=poisson_ref,
        moment_corrections=moment_corrections)

    rep = repr(poisson)
    expected_repr = f'MomentCorrectionPoissonSolver ({expected_string})'

    assert isinstance(rep, str)
    assert rep == expected_repr, f'{rep} not equal to {expected_repr}'


@pytest.fixture
def gd():
    N_c = (16, 16, 3 * 16)
    cell_cv = (1, 1, 3)
    gd = GridDescriptor(N_c, cell_cv, False)

    return gd


@pytest.mark.parametrize('moment_corrections', [
    4,
    9,
    [dict(moms=range(4), center=np.array([1, 1, 1]))],
    [dict(moms=range(4), center=np.array([2, 3, 4])),
     dict(moms=range(4), center=np.array([7.4, 3.1, 0.1]))],
])
def test_write(gd, moment_corrections):
    poisson_ref = PoissonSolver()
    poisson_ref.set_grid_descriptor(gd)

    poisson = MomentCorrectionPoissonSolver(
        poissonsolver=poisson_ref,
        moment_corrections=moment_corrections)
    poisson.set_grid_descriptor(gd)

    from gpaw.io import Writer
    from gpaw.mpi import world
    filename = '/dev/null'

    # By using the Writer we check that everything is JSON serializable
    writer = Writer(filename, world)
    writer.child('poisson').write(**poisson.todict())
    writer.close()


@pytest.fixture
def rho_g(gd):
    # Construct model density
    coord_vg = gd.get_grid_point_coordinates()
    z_g = coord_vg[2, :]
    rho_g = gd.zeros()
    for z0 in [1, 2]:
        rho_g += 10 * (z_g - z0) * \
            np.exp(-20 * np.sum((coord_vg.T - np.array([.5, .5, z0])).T**2,
                                axis=0))

    return rho_g


@pytest.fixture
def poisson_solve(gd, rho_g):

    def _poisson_solve(poisson):
        poisson.set_grid_descriptor(gd)
        phi_g = gd.zeros()
        poisson.solve(phi_g, rho_g)

        return phi_g

    return _poisson_solve


@pytest.fixture
def compare(gd, tolerance, cmp_begin):
    # Some test cases compare in only a small region of space
    if cmp_begin is None:
        slice = None
    else:
        Ng_c = gd.get_size_of_global_array()
        cmp_end = 1 - cmp_begin
        idx_c = [np.arange(int(N * cmp_begin), int(N * cmp_end)) for N in Ng_c]
        slice = np.ix_(*idx_c)

    def _compare(phi1_g, phi2_g):
        big_phi1_g = gd.collect(phi1_g)
        big_phi2_g = gd.collect(phi2_g)
        if gd.comm.rank == 0:
            if slice is not None:
                big_phi1_g = big_phi1_g[slice]
                big_phi2_g = big_phi2_g[slice]
            equal(np.max(np.absolute(big_phi1_g - big_phi2_g)), 0.0, tolerance)

    return _compare


@pytest.fixture
def poisson_ref(gd, ref):
    poisson_default = PoissonSolver()
    if ref == 'default':
        # Get reference from default poissonsolver
        # Using the default solver the potential is forced to zero at the box
        # boundries. The potential thus has the wrong shape near the boundries
        # but is nearly right in the center of the box
        return poisson_default
    elif ref == 'extravac':
        # Get reference from extravacuum solver
        # With 4 times extra vacuum the potential is well converged everywhere
        poisson_extravac = ExtraVacuumPoissonSolver(
            gpts=4 * gd.N_c,
            poissonsolver_large=poisson_default)
        return poisson_extravac
    else:
        raise ValueError(f'No such ref {ref}')


@pytest.mark.parametrize('ref, moment_corrections, tolerance, cmp_begin', [
    # MomentCorrectionPoissonSolver without any moment corrections should be
    # exactly as the underlying solver
    ('default', None, 0.0, None),
    # It should also be possible to chain default+extravacuum+moment correction
    # With moment_correction=None the MomentCorrection solver doesn't actually
    # do anything, so the potential should be identical to the extra vacuum
    # reference
    ('extravac', None, 0.0, None),
    # Test moment_corrections=int
    # The moment correction is applied to the center of the cell. This is not
    # enough to have a converged potential near the edges
    # The closer we are to the center the better though
    ('default', 4, 3.5e-2, 0.25),
    ('default', 4, 2.5e-2, 0.40),
    # Test moment_corrections=list
    # Remember that the solver expects Ångström units and we have specified
    # the grid in Bohr
    # This should give a well converged potential everywhere, that we can
    # compare to the reference extravacuum potential
    ('extravac',
     [{'moms': range(4), 'center': np.array([.5, .5, 1]) * Bohr},
      {'moms': range(4), 'center': np.array([.5, .5, 2]) * Bohr}],
     3e-3, None),
    # It should be possible to chain default+extravacuum+moment correction
    # As the potential is already well converged, there should be little change
    ('extravac',
     [{'moms': range(4), 'center': np.array([.5, .5, 1]) * Bohr},
      {'moms': range(4), 'center': np.array([.5, .5, 2]) * Bohr}],
     5e-4, None),
])
def test_poisson_moment_correction(gd, rho_g, poisson_solve,
                                   compare, poisson_ref,
                                   ref, moment_corrections,
                                   tolerance, cmp_begin):
    # Solve for the potential using the reference solver
    phiref_g = poisson_solve(poisson_ref)

    # Create a MomentCorrectionPoissonSolver and solve for the potential
    poisson = MomentCorrectionPoissonSolver(poissonsolver=poisson_ref,
                                            moment_corrections=None)
    phi_g = poisson_solve(poisson)

    # Test the MomentCorrectionPoissonSolver
    compare(phi_g, phiref_g)



import pytest
from ase.io import read


@pytest.mark.skip
def test_yaml(gpw_files):
    pytest.importorskip('yaml')
    print(gpw_files['h2_pw'].with_name('h2_pw.txt'))
    a = read(gpw_files['h2_pw'].with_name('h2_pw.txt'))
    print(a)



import pytest
import numpy as np
from ase.units import Bohr
from ase.build import bulk
from gpaw import GPAW, FermiDirac
from gpaw.response.df import DielectricFunction
from gpaw.test import equal, findpeak


@pytest.mark.response
@pytest.mark.libxc
def test_response_diamond_absorption(in_tmp_dir):
    a = 6.75 * Bohr
    atoms = bulk('C', 'diamond', a=a)

    calc = GPAW(mode='pw',
                kpts=(3, 3, 3),
                eigensolver='rmm-diis',
                occupations=FermiDirac(0.001))

    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('C.gpw', 'all')

    eM1_ = 9.727
    eM2_ = 9.548
    w0_ = 10.7782
    I0_ = 5.47
    w_ = 10.7532
    I_ = 5.98

    # Macroscopic dielectric constant calculation
    df = DielectricFunction('C.gpw', frequencies=(0.,), eta=0.001, ecut=50,
                            hilbert=False)
    eM1, eM2 = df.get_macroscopic_dielectric_constant()
    equal(eM1, eM1_, 0.01)
    equal(eM2, eM2_, 0.01)

    # Absorption spectrum calculation RPA
    df = DielectricFunction('C.gpw',
                            eta=0.25,
                            ecut=50,
                            frequencies=np.linspace(0, 24., 241),
                            hilbert=False)
    a0, a = df.get_dielectric_function(filename=None)
    df.check_sum_rule(a.imag)

    equal(a0[0].real, eM1_, 0.01)
    equal(a[0].real, eM2_, 0.01)
    w, I = findpeak(np.linspace(0, 24., 241), a0.imag)
    equal(w, w0_, 0.01)
    equal(I / (4 * np.pi), I0_, 0.1)
    w, I = findpeak(np.linspace(0, 24., 241), a.imag)
    equal(w, w_, 0.01)
    equal(I / (4 * np.pi), I_, 0.1)

    a0, a = df.get_polarizability(filename=None)

    w, I = findpeak(np.linspace(0, 24., 241), a0.imag)
    equal(w, w0_, 0.01)
    equal(I, I0_, 0.01)
    w, I = findpeak(np.linspace(0, 24., 241), a.imag)
    equal(w, w_, 0.01)
    equal(I, I_, 0.01)

    # Absorption spectrum calculation ALDA
    w0_ = 10.7931
    I0_ = 5.36
    w_ = 10.7562
    I_ = 5.8803

    df.get_polarizability(xc='ALDA', filename='ALDA_pol.csv')
    # Here we base the check on a written results file
    df.context.comm.barrier()
    d = np.loadtxt('ALDA_pol.csv', delimiter=',')

    w, I = findpeak(d[:, 0], d[:, 2])
    equal(w, w0_, 0.01)
    equal(I, I0_, 0.1)
    w, I = findpeak(d[:, 0], d[:, 4])
    equal(w, w_, 0.01)
    equal(I, I_, 0.1)

    # Absorption spectrum calculation long-range kernel
    w0_ = 10.2189
    I0_ = 5.14
    w_ = 10.2906
    I_ = 5.6955

    a0, a = df.get_polarizability(filename=None, xc='LR0.25')

    w, I = findpeak(np.linspace(0, 24., 241), a0.imag)
    equal(w, w0_, 0.01)
    equal(I, I0_, 0.1)
    w, I = findpeak(np.linspace(0, 24., 241), a.imag)
    equal(w, w_, 0.01)
    equal(I, I_, 0.1)

    # Absorption spectrum calculation Bootstrap
    w0_ = 10.37
    I0_ = 5.27
    w_ = 10.4600
    I_ = 6.0263

    a0, a = df.get_polarizability(filename=None, xc='Bootstrap')

    w, I = findpeak(np.linspace(0, 24., 241), a0.imag)
    equal(w, w0_, 0.02)
    equal(I, I0_, 0.2)
    w, I = findpeak(np.linspace(0, 24., 241), a.imag)
    equal(w, w_, 0.02)
    equal(I, I_, 0.2)



import pytest
from gpaw.mpi import world
import numpy as np
from gpaw.response.df import DielectricFunction
from gpaw.response.bse import BSE, read_spectrum
from gpaw.test import findpeak, equal

pytestmark = pytest.mark.skipif(world.size < 4,
                                reason='world.size < 4')


@pytest.mark.response
def test_response_bse_aluminum(in_tmp_dir, gpw_files):
    df = 1
    bse = 1
    check_spectrum = 1

    q_c = np.array([0.25, 0.0, 0.0])
    w_w = np.linspace(0, 24, 241)
    eta = 0.2
    ecut = 50
    if bse:
        bse = BSE(gpw_files['bse_al'],
                  valence_bands=range(4),
                  conduction_bands=range(4),
                  mode='RPA',
                  nbands=4,
                  ecut=ecut,
                  write_h=False,
                  write_v=False,
                  )
        bse.get_eels_spectrum(filename='bse_eels.csv',
                              q_c=q_c,
                              w_w=w_w,
                              eta=eta)
        omega_w, bse_w = read_spectrum('bse_eels.csv')

    if df:
        df = DielectricFunction(calc=gpw_files['bse_al'],
                                frequencies=w_w,
                                eta=eta,
                                ecut=ecut,
                                hilbert=False)
        df_w = df.get_eels_spectrum(q_c=q_c, filename=None)[1]

    if check_spectrum:
        assert w_w == pytest.approx(omega_w)
        w_ = 15.1423
        I_ = 25.4359
        wbse, Ibse = findpeak(w_w, bse_w)
        wdf, Idf = findpeak(w_w, df_w)
        equal(wbse, w_, 0.01)
        equal(wdf, w_, 0.01)
        equal(Ibse, I_, 0.1)
        equal(Idf, I_, 0.1)



import numpy as np
import numpy.random as ra
from gpaw.test import equal
from gpaw.setup import create_setup
from gpaw.grid_descriptor import GridDescriptor
from gpaw.lfc import LFC
from gpaw.spline import Spline
from gpaw.xc import XC
from gpaw.utilities import pack
from gpaw.mpi import serial_comm


def test_xc_gga_atom():
    ra.seed(8)
    for name in ['LDA', 'PBE']:
        xc = XC(name)
        s = create_setup('N', xc)
        ni = s.ni
        nao = s.nao
        wt0_j = s.basis_functions_J

        rcut = s.xc_correction.rgd.r_g[-1]

        wt_j = []
        for wt0 in wt0_j:
            data = [wt0(r) for r in np.arange(121) * rcut / 100]
            data[-1] = 0.0
            l = wt0.get_angular_momentum_number()
            wt_j.append(Spline(l, 1.2 * rcut, data))

        a = rcut * 1.2 * 2 + 1.0
        n = 70
        n = 90
        gd = GridDescriptor((n, n, n), (a, a, a), comm=serial_comm)
        pr = LFC(gd, [wt_j])
        pr.set_positions([(0.5, 0.5, 0.5)])

        psit_ig = np.zeros((nao, n, n, n))
        pr.add(psit_ig, {0: np.eye(nao)})

        nii = ni * (ni + 1) // 2
        D_p = np.zeros(nii)

        e_g = np.zeros((n, n, n))
        n_g = np.zeros((1, n, n, n))
        v_g = np.zeros((1, n, n, n))

        P_ni = 0.2 * ra.random((20, ni))
        P_ni[:, nao:] = 0.0
        D_ii = np.dot(np.transpose(P_ni), P_ni)
        D_p = pack(D_ii)
        p = 0
        for i1 in range(nao):
            for i2 in range(i1, nao):
                n_g += D_p[p] * psit_ig[i1] * psit_ig[i2]
                p += 1
            p += ni - nao

        p = LFC(gd, [[s.nct]])
        p.set_positions([(0.5, 0.5, 0.5)])
        p.add(n_g[0], 1.0)
        e_g = gd.zeros()
        xc.calculate(gd, n_g, v_g, e_g)

        r2_g = np.sum((np.indices((n, n, n)) - n / 2)**2, axis=0)
        dv_g = gd.dv * np.less(r2_g, (rcut / a * n)**2)

        E2 = -np.dot(e_g.ravel(), dv_g.ravel())

        s.xc_correction.n_qg[:] = 0.0
        s.xc_correction.nc_g[:] = 0.0
        E1 = (xc.calculate_paw_correction(s, D_p.reshape(1, -1))
              + s.xc_correction.e_xc0)

        print(name, E1, E2, E1 - E2)
        equal(E1, E2, 0.0013)



import numpy as np
import pytest

from ase.parallel import world, parprint
from ase.units import Bohr
from gpaw import GPAW
from gpaw.lcao.dipoletransition import get_dipole_transitions
from gpaw.utilities.dipole import dipole_matrix_elements_from_calc
from gpaw.lrtddft.kssingle import KSSingles


@pytest.mark.later
def test_dipole_transition(gpw_files, tmp_path_factory):
    """Check dipole matrix-elements for H20."""
    calc = GPAW(gpw_files['h2o_lcao'])
    # Initialize calculator if necessary
    if not hasattr(calc.wfs, 'C_nM'):
        calc.initialize_positions(calc.atoms)
    dip_skvnm = get_dipole_transitions(calc.wfs).real
    parprint("Dipole moments calculated")
    assert dip_skvnm.shape == (1, 1, 3, 6, 6)
    dip_vnm = dip_skvnm[0, 0] * Bohr

    print(world.rank, dip_vnm[0, 0, 3])

    # check symmetry: abs(d[i,j]) == abs(d[j,i])
    for v in range(3):
        dip_vnm[v].T == pytest.approx(dip_vnm[v])

    # Check numerical value of a few elements - signs might change!
    assert 0.0693 == pytest.approx(abs(dip_vnm[2, 0, 4]), abs=1e-4)
    assert 0.1014 == pytest.approx(abs(dip_vnm[1, 0, 5]), abs=1e-4)
    assert 0.1709 == pytest.approx(abs(dip_vnm[0, 3, 4]), abs=1e-4)

    # some printout for manual inspection, if wanted
    f = 6 * "{:+.4f} "
    for c in range(3):
        for i in range(6):
            parprint(f.format(*dip_vnm[c, i]))
        parprint("")

    # ------------------------------------------------------------------------
    # compare to utilities implementation
    if world.rank == 0:
        from gpaw.new.ase_interface import GPAW as NewGPAW
        from gpaw.mpi import serial_comm
        refcalc = NewGPAW(gpw_files['h2o_lcao'],
                          communicator=serial_comm)
        uref = dipole_matrix_elements_from_calc(refcalc, 0, 6)
        uref = uref[0]
        assert uref.shape == (6, 6, 3)
    # NOTE: Comparing implementations of r gauge and v gauge is tricky, as they
    # tend to be numerically inequivalent.

    # compare to lrtddft implementation
    kss = KSSingles()
    atoms = calc.atoms
    atoms.calc = calc
    kss.calculate(calc.atoms, 1)
    lrref = []
    lrrefv = []
    for ex in kss:
        lrref.append(-1. * ex.mur * Bohr)
        lrrefv.append(-1. * ex.muv * Bohr)
    lrref = np.array(lrref)
    lrrefv = np.array(lrrefv)

    # Additional benefit: tests equivalence of r gauge implementations
    if world.rank == 0:
        for i, (m, n, v) in enumerate([[4, 0, 2],
                                       [5, 0, 1],
                                       [4, 1, 1],
                                       [5, 1, 2],
                                       [4, 2, 2],
                                       [5, 2, 1],
                                       [4, 3, 0]]):
            assert abs(lrref[i, v]) == pytest.approx(abs(uref[m, n, v]),
                                                     abs=1e-4)

    # some printout for manual inspection, if wanted
    parprint("         r-gauge   lrtddft(v)  raman(v)")
    f = "{} {:+.4f}    {:+.4f}    {:+.4f}"
    parprint(f.format('0->4 (z)', lrref[0, 2], lrrefv[0, 2], dip_vnm[2, 0, 4]))
    parprint(f.format('0->5 (y)', lrref[1, 1], lrrefv[1, 1], dip_vnm[1, 0, 5]))
    parprint(f.format('1->4 (y)', lrref[2, 1], lrrefv[2, 1], dip_vnm[1, 1, 4]))
    parprint(f.format('1->5 (z)', lrref[3, 2], lrrefv[3, 2], dip_vnm[2, 1, 5]))
    parprint(f.format('2->4 (z)', lrref[4, 2], lrrefv[4, 2], dip_vnm[2, 2, 4]))
    parprint(f.format('2->5 (y)', lrref[5, 1], lrrefv[5, 1], dip_vnm[1, 2, 5]))
    parprint(f.format('3->4 (x)', lrref[6, 0], lrrefv[6, 0], dip_vnm[0, 3, 4]))



import pytest
from gpaw.grid_descriptor import GridDescriptor
from gpaw import PoissonSolver
from gpaw.spline import Spline
from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.response.pair_functions import SingleQPWDescriptor
import numpy as np
from gpaw.response.coulomb_kernels import get_coulomb_kernel
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.poisson_extravacuum import ExtraVacuumPoissonSolver
from gpaw.poisson_moment import MomentCorrectionPoissonSolver


class ExtraVacuum2DPoisson:
    def __init__(self, poisson):
        self.poisson = poisson

    def set_grid_descriptor(self, gd):
        self.N_c = gd.N_c
        self.gd = GridDescriptor(gd.N_c * np.array([1, 1, 9]),
                                 gd.cell_cv @ np.diag([1, 1, 9]),
                                 pbc_c=gd.pbc_c)
        self.poisson.set_grid_descriptor(self.gd)

    def solve(self, v, n):
        myv = self.gd.zeros()
        myn = self.gd.zeros()
        beg, end = self.N_c[2] * 4, self.N_c[2] * 5 - 1
        myn[:, :, beg:end] = n
        self.poisson.solve(myv, myn)
        v[:] = myv[:, :, beg:end]


class Grid:
    def __init__(self, L, N, q, truncation):
        if q == 0:
            supercell = 1
        else:
            supercell = round(1 / q)
            assert np.allclose(supercell * q, 1.0)

        # First we create a unit cell quantities
        pbc_c = [True, True, True]
        if truncation is None:
            pass
        elif truncation == '0D':
            pbc_c = [False, False, False]
        elif truncation == '2D':
            pbc_c = [True, True, False]
        else:
            raise NotImplementedError
        pbc_c = np.array(pbc_c)
        self.pbc_c = pbc_c
        gd = GridDescriptor((N, N, N), [L, L, L], pbc_c=pbc_c)
        self.periodicgd = GridDescriptor((N, N, N), [L, L, L])
        kd = KPointDescriptor([[q, 0, 0]])
        # Create some charge distribution to the unit cell
        if truncation == '0D':
            cut = L / 9
        else:
            cut = L / 4
        spline = Spline(l=0, rmax=cut, f_g=np.array([1, 0.5, 0.0]))
        c = LFC(gd, [[spline]] * 2, kd=kd, dtype=complex)

        if truncation == '0D':
            c.set_positions([[0.5 - 0.1, 0.5, 0.5], [0.5 + 0.1, 0.5, 0.5]])
        else:
            c.set_positions([[0.11, 0.0, 0.5], [0.5, 0.39, 0.5]])
        n_R = gd.zeros(dtype=complex)
        c.add(n_R, {0: np.array([-1], dtype=complex),
                    1: np.array([1], dtype=complex)}, 0)

        # Now we replicate to supercell, to calculate using grid based Poisson
        # solvers which cannot handle Bloch-phases at the moment
        supergd = GridDescriptor((N * supercell, N, N),
                                 [L * supercell, L, L],
                                 pbc_c=pbc_c)

        # Avoid going over 79 characters
        EVPS = ExtraVacuumPoissonSolver
        MCPS = MomentCorrectionPoissonSolver
        PS = PoissonSolver
        if truncation is None:
            superpoisson = PS(nn=3)
        elif truncation == '0D':
            mcps = MCPS(PS(nn=3), moment_corrections=9)
            superpoisson = EVPS(gpts=(N * 4, N * 4, N * 4),
                                poissonsolver_large=mcps,
                                coarses=1,
                                poissonsolver_small=PS(nn=3))
        elif truncation == '2D':
            superpoisson = ExtraVacuum2DPoisson(PS(nn=3))
        else:
            raise NotImplementedError(truncation)
        superpoisson.set_grid_descriptor(supergd)

        if q != 0:
            supern_R = supergd.zeros(dtype=complex)
            for i in range(supercell):
                pn_R = n_R * np.exp(2j * np.pi * i / supercell)
                supern_R[(i * N):((i + 1) * N), :, :] = pn_R

        else:
            supergd = gd
            supern_R = n_R

        superv_R = supergd.zeros(dtype=complex)
        superpoisson.solve(superv_R.real, supern_R.real)
        superpoisson.solve(superv_R.imag, supern_R.imag)

        self.Ec = supergd.integrate(supern_R, superv_R) / supercell

        self.gd = gd
        self.n_R = gd.zero_pad(n_R)


@pytest.mark.serial
@pytest.mark.response
@pytest.mark.parametrize('gridparam', [(32, 1e-5), (64, 1e-7), (96, 1e-8)])
@pytest.mark.parametrize('qtrunc', [(0, '2D'),
                                    (1 / 3, '2D'),
                                    (0, None),
                                    (1 / 3, None),
                                    (0, '0D')])
def test_coulomb(gridparam, qtrunc):
    N, maxdev = gridparam
    q, truncation = qtrunc
    L = 10
    print()
    # Slightly lower tolerance for 0D system, because it uses so localized
    # charges to avoid crosstalk
    if truncation == '0D':
        maxdev *= 1e2
    grid = Grid(L, N, q, truncation=truncation)
    # Use maximum ecut
    ecut0 = 0.5 * np.pi**2 / ((L / N)**2)
    qpd = SingleQPWDescriptor.from_q([q, 0, 0], ecut0, grid.periodicgd,
                                     gammacentered=False)
    # XXX It is silly that get_coulomb_kernel uses k-point numbers
    # per dim to determine non-periodic direction. It should just get
    # non_per_dir=2, etc.
    v_G = get_coulomb_kernel(qpd, np.array([2, 2, 1]), truncation=truncation,
                             pbc_c=grid.pbc_c)
    n_G = qpd.fft(grid.n_R * grid.periodicgd.plane_wave([-q, 0, 0]))
    Ec2 = n_G.conj() @ (v_G * n_G) / N**6 * L**3
    dev = np.abs(grid.Ec / Ec2 - 1)
    print(f'N={N:4d} q=[{q:.3f},0,0] truncation: {str(truncation):5s}'
          f'rel. log10 deviation: {np.log10(dev):7.4f}'
          f'coulomb:{Ec2:9.5f}')
    assert np.abs(dev) < maxdev



import numpy as np


def get_fHxc_Gr(xcflags, rs, q, qF, s2_g):
    xc = xcflags.xc
    if xc == 'rALDA':
        return fHxc_ralda(q, qF)

    elif xcflags.is_apbe:
        return fHxc_apbe(rs, q, s2_g)
    else:
        raise ValueError(f'Unknown xc: {xc}')


def fHxc_ralda(q, qF):
    # rALDA (exchange only) kernel
    # Olsen and Thygesen, Phys. Rev. B 88, 115131 (2013)
    # ALDA up to 2*qF, -vc for q >2qF (such that fHxc vanishes)

    rxalda_A = 0.25
    rxalda_qcut = qF * np.sqrt(1.0 / rxalda_A)

    # construct fHxc(k,r)
    return (0.5 + 0.0j) * (
        (1.0 + np.sign(rxalda_qcut - q[:, np.newaxis])) *
        (1.0 + (-1.0) * rxalda_A * (q[:, np.newaxis] / qF)**2.0))


def fHxc_apbe(rs, q, s2_g):
    # Olsen and Thygesen, Phys. Rev. Lett. 112, 203001 (2014)
    # Exchange only part of the PBE XC kernel, neglecting the terms
    # arising from the variation of the density gradient
    # i.e. second functional derivative
    # d2/drho^2 -> \partial^2/\partial rho^2 at fixed \nabla \rho

    fxc_PBE = get_pbe_fxc(
        pbe_rho=3.0 / (4.0 * np.pi * rs**3.0),
        pbe_s2_g=s2_g)
    rxapbe_qcut = np.sqrt(-4.0 * np.pi / fxc_PBE)

    return (0.5 + 0.0j) * (
        (1.0 + np.sign(rxapbe_qcut - q[:, np.newaxis])) *
        (1.0 + fxc_PBE / (4.0 * np.pi) * (q[:, np.newaxis])**2.0))


def get_fspinHxc_Gr_rALDA(qF, q):
    rxalda_A = 0.25
    rxalda_qcut = qF * np.sqrt(1.0 / rxalda_A)

    fspinHxc_Gr = ((0.5 + 0.0j) *
                   (1.0 + np.sign(rxalda_qcut - q[:, np.newaxis])) *
                   (-1.0) * rxalda_A * (q[:, np.newaxis] / qF)**2.0)
    return fspinHxc_Gr


def get_fspinHxc_Gr_rAPBE(rs, q, s2_g):
    fxc_PBE = get_pbe_fxc(pbe_rho=(3.0 / (4.0 * np.pi * rs**3.0)),
                          pbe_s2_g=s2_g)
    rxapbe_qcut = np.sqrt(-4.0 * np.pi / fxc_PBE)
    fspinHxc_Gr = ((0.5 + 0.0j) *
                   (1.0 + np.sign(rxapbe_qcut - q[:, np.newaxis])) *
                   fxc_PBE / (4.0 * np.pi) * q[:, np.newaxis]**2.0)
    return fspinHxc_Gr


def get_heg_A(rs):
    # Returns the A coefficient, where the
    # q ->0 limiting value of static fxc
    # of the HEG = -\frac{4\pi A }{q_F^2} = f_{xc}^{ALDA}.
    # We need correlation energy per electron and first and second derivs
    # w.r.t. rs
    # See for instance Moroni, Ceperley and Senatore,
    # Phys. Rev. Lett. 75, 689 (1995)
    # (and also Kohn and Sham, Phys. Rev. 140, A1133 (1965) equation 2.7)

    # Exchange contribution
    heg_A = 0.25

    # Correlation contribution
    A_ec, A_dec, A_d2ec = get_pw_lda(rs)
    heg_A += (1.0 / 27.0 * rs**2.0 * (9.0 * np.pi / 4.0)**(2.0 / 3.0) *
              (2 * A_dec - rs * A_d2ec))

    return heg_A


def get_heg_B(rs):
    # Returns the B coefficient, where the
    # q -> oo limiting behaviour of static fxc
    # of the HEG is -\frac{4\pi B}{q^2} - \frac{4\pi C}{q_F^2}.
    # Use the parametrisation of Moroni, Ceperley and Senatore,
    # Phys. Rev. Lett. 75, 689 (1995)

    mcs_xs = np.sqrt(rs)

    mcs_a = (1.0, 2.15, 0.0, 0.435)
    mcs_b = (3.0, 1.57, 0.0, 0.409)

    mcs_num = 0

    for mcs_j, mcs_coeff in enumerate(mcs_a):
        mcs_num += mcs_coeff * mcs_xs**mcs_j

    mcs_denom = 0

    for mcs_j, mcs_coeff in enumerate(mcs_b):
        mcs_denom += mcs_coeff * mcs_xs**mcs_j

    heg_B = mcs_num / mcs_denom
    return heg_B


def get_heg_C(rs):
    # Returns the C coefficient, where the
    # q -> oo limiting behaviour of static fxc
    # of the HEG is -\frac{4\pi B}{q^2} - \frac{4\pi C}{q_F^2}.
    # Again see Moroni, Ceperley and Senatore,
    # Phys. Rev. Lett. 75, 689 (1995)

    C_ec, C_dec, Cd2ec = get_pw_lda(rs)

    heg_C = ((-1.0) * np.pi**(2.0 / 3.0) * (1.0 / 18.0)**(1.0 / 3.0) *
             (rs * C_ec + rs**2.0 * C_dec))

    return heg_C


def get_heg_D(rs):
    # Returns a 'D' coefficient, where the
    # q->0 omega -> oo limiting behaviour
    # of the frequency dependent fxc is -\frac{4\pi D}{q_F^2}
    # see Constantin & Pitarke Phys. Rev. B 75, 245127 (2007) equation 7

    D_ec, D_dec, D_d2ec = get_pw_lda(rs)

    # Exchange contribution
    heg_D = 0.15
    # Correlation contribution
    heg_D += ((9.0 * np.pi / 4.0)**(2.0 / 3.0) * rs / 3.0 *
              (22.0 / 15.0 * D_ec + 26.0 / 15.0 * rs * D_dec))
    return heg_D


def get_pw_lda(rs):
    # Returns LDA correlation energy and its first and second
    # derivatives with respect to rs according to the parametrisation
    # of Perdew and Wang, Phys. Rev. B 45, 13244 (1992)

    pw_A = 0.031091
    pw_alp = 0.21370
    pw_beta = (7.5957, 3.5876, 1.6382, 0.49294)

    pw_logdenom = 2.0 * pw_A * (
        pw_beta[0] * rs**0.5 + pw_beta[1] * rs**1.0 +
        pw_beta[2] * rs**1.5 + pw_beta[3] * rs**2.0)

    pw_dlogdenom = 2.0 * pw_A * (0.5 * pw_beta[0] * rs**(-0.5) +
                                 1.0 * pw_beta[1] +
                                 1.5 * pw_beta[2] * rs**0.5 +
                                 2.0 * pw_beta[3] * rs)

    pw_d2logdenom = 2.0 * pw_A * (-0.25 * pw_beta[0] * rs**(-1.5) +
                                  0.75 * pw_beta[2] * rs**(-0.5) +
                                  2.0 * pw_beta[3])

    pw_logarg = 1.0 + 1.0 / pw_logdenom
    pw_dlogarg = (-1.0) / (pw_logdenom**2.0) * pw_dlogdenom
    pw_d2logarg = 2.0 / (pw_logdenom**3.0) * (pw_dlogdenom**2.0)
    pw_d2logarg += (-1.0) / (pw_logdenom**2.0) * pw_d2logdenom

    # pw_ec = the correlation energy (per electron)
    pw_ec = -2.0 * pw_A * (1 + pw_alp * rs) * np.log(pw_logarg)

    # pw_dec = first derivative

    pw_dec = -2.0 * pw_A * (1 + pw_alp * rs) * pw_dlogarg / pw_logarg
    pw_dec += (-2.0 * pw_A * pw_alp) * np.log(pw_logarg)

    # pw_d2ec = second derivative

    pw_d2ec = (-2.0) * pw_A * pw_alp * pw_dlogarg / pw_logarg
    pw_d2ec += (-2.0) * pw_A * ((1 + pw_alp * rs) *
                                (pw_d2logarg / pw_logarg -
                                 (pw_dlogarg**2.0) / (pw_logarg**2.0)))
    pw_d2ec += (-2.0 * pw_A * pw_alp) * pw_dlogarg / pw_logarg

    return pw_ec, pw_dec, pw_d2ec


def get_pbe_fxc(pbe_rho, pbe_s2_g):
    return get_pbe_fxc_and_intermediate_derivatives(pbe_rho, pbe_s2_g)[0]


def get_pbe_fxc_and_intermediate_derivatives(pbe_rho, pbe_s2_g):
    # The intermediate derivatives are only used for testing.
    pbe_kappa = 0.804
    pbe_mu = 0.2195149727645171

    pbe_denom_g = 1.0 + pbe_mu * pbe_s2_g / pbe_kappa

    F_g = 1.0 + pbe_kappa - pbe_kappa / pbe_denom_g
    Fn_g = -8.0 / 3.0 * pbe_mu * pbe_s2_g / pbe_rho / pbe_denom_g**2.0
    Fnn_g = (-11.0 / 3.0 / pbe_rho * Fn_g -
             2.0 / pbe_kappa * Fn_g**2.0 * pbe_denom_g)

    e_g = -3.0 / 4.0 * (3.0 / np.pi)**(1.0 / 3.0) * pbe_rho**(4.0 / 3.0)
    v_g = 4.0 / 3.0 * e_g / pbe_rho
    f_g = 1.0 / 3.0 * v_g / pbe_rho

    pbe_f_g = f_g * F_g + 2.0 * v_g * Fn_g + e_g * Fnn_g
    return pbe_f_g, F_g, Fn_g, Fnn_g
